from typing import Optional, Union

from textx import TextXSyntaxError

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.document_view import (
    DocumentView,
    ViewElement,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementField,
)
from strictdoc.backend.sdoc.models.node import (
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.reference import Reference
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    ForwardFunctionRangeMarker,
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    ForwardRangeMarker,
    RangeMarker,
)
from strictdoc.helpers.exception import StrictDocException


def get_textx_syntax_error_message(exception: TextXSyntaxError) -> str:
    return f"SDoc markup error: {exception.context}."


class StrictDocSemanticError(StrictDocException):
    def __init__(
        self,
        title: str,
        hint: Optional[str],
        example: Optional[str],
        line: Optional[int] = None,
        col: Optional[int] = None,
        filename: Optional[str] = None,
    ) -> None:
        super().__init__(title, hint, line, col, filename)
        self.title = title
        self.hint = hint
        self.example = example
        self.line = line
        self.col = col
        self.file_path = filename

    @staticmethod
    def unknown_node_type(
        node: SDocNode, path_to_sdoc_file: str
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=f"Invalid node type: {node.node_type}.",
            hint=None,
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def composite_node_and_non_composite_element_mismatch(
        node: SDocNode, path_to_sdoc_file: str
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=f"The composite node's grammar element is declared as non-composite: [[{node.node_type}]].",
            hint=(
                "The composite node grammar element declaration must "
                "have a 'PROPERTIES/IS_COMPOSITE: True' declaration."
            ),
            example="""\
[GRAMMAR]
ELEMENTS:
- TAG: NODE
  PROPERTIES:
    IS_COMPOSITE: True
  FIELDS:
  ...
""",
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def non_composite_node_and_composite_element_composite(
        node: SDocNode, path_to_sdoc_file: str
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=f"The non-composite node's grammar element is declared as composite: [[{node.node_type}]].",
            hint=(
                "The composite node grammar element declaration must "
                "have a 'PROPERTIES/IS_COMPOSITE: False' declaration or the "
                "PROPERTIES/IS_COMPOSITE can be simply omitted."
            ),
            example="""\
[GRAMMAR]
ELEMENTS:
- TAG: NODE
  PROPERTIES:
    IS_COMPOSITE: False
  FIELDS:
  ...
""",
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def unregistered_field(
        *,
        field_name: str,
        requirement: SDocNode,
        document_grammar: DocumentGrammar,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        grammar_dump = document_grammar.dump_fields(requirement.node_type)
        return StrictDocSemanticError(
            title=f"Invalid requirement field: {field_name}",
            hint=(
                f"Compare with the document grammar: [{grammar_dump}] "
                f"for type: {requirement.node_type}."
            ),
            example=None,
            line=requirement.ng_line_start,
            col=requirement.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def missing_required_field(
        node: SDocNode,
        grammar_field: GrammarElementField,
        document_grammar: DocumentGrammar,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        grammar_fields = document_grammar.dump_fields(node.node_type)
        return StrictDocSemanticError(
            title=(
                f"Node is missing a field that is required by "
                f"grammar: {grammar_field.title}."
            ),
            hint=(
                f"Node fields: [{node.dump_fields_as_parsed()}], "
                f"grammar fields: [{grammar_fields}]."
            ),
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def unexpected_field_outside_grammar(
        node: SDocNode,
        requirement_field: SDocNodeField,
        document_grammar: DocumentGrammar,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        grammar_fields = document_grammar.dump_fields(node.node_type)
        return StrictDocSemanticError(
            title=(
                f"Unexpected field outside grammar: "
                f"{requirement_field.field_name}"
            ),
            hint=(
                f"Requirement fields: [{node.dump_fields_as_parsed()}], "
                f"grammar fields: [{grammar_fields}]."
            ),
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def wrong_field_order(
        node: SDocNode,
        document_grammar: DocumentGrammar,
        problematic_field: SDocNodeField,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        assert isinstance(problematic_field, SDocNodeField), (
            f"{problematic_field}"
        )
        requirement_dump = node.dump_fields_as_parsed()
        grammar_dump = document_grammar.dump_fields(node.node_type)
        return StrictDocSemanticError(
            title=f"Wrong field order for requirement: [{requirement_dump}].",
            hint=(
                f"Problematic field: {problematic_field.field_name}. "
                f"Compare with the document grammar: [{grammar_dump}] "
                f"for type: {node.node_type}."
            ),
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def invalid_choice_field(
        node: SDocNode,
        document_grammar: DocumentGrammar,
        requirement_field: SDocNodeField,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=(
                f"Requirement field has an invalid SingleChoice value: "
                f"{requirement_field.get_text_value()}."
            ),
            hint=(
                f"Problematic field: {requirement_field.field_name}. "
                f"Compare with the document grammar: "
                f"["
                f"{document_grammar.dump_fields(node.node_type)}"
                f"] "
                f"for type: {node.node_type}."
            ),
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def invalid_multiple_choice_field(
        node: SDocNode,
        document_grammar: DocumentGrammar,
        requirement_field: SDocNodeField,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=(
                f"Requirement field has an invalid MultipleChoice value: "
                f"{requirement_field.get_text_value()}."
            ),
            hint=(
                f"Problematic field: {requirement_field.field_name}. "
                f"Compare with the document grammar: "
                f"["
                f"{document_grammar.dump_fields(node.node_type)}"
                f"] "
                f"for type: {node.node_type}."
            ),
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def not_comma_separated_choices(
        node: SDocNode,
        requirement_field: SDocNodeField,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=(
                f"Requirement field of type MultipleChoice is invalid: "
                f"{requirement_field.get_text_value()}."
            ),
            hint="MultipleChoice field requires ', '-separated values.",
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def not_comma_separated_tag_field(
        node: SDocNode,
        requirement_field: SDocNodeField,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=(
                f"Requirement field of type Tag is invalid: "
                f"{requirement_field.get_text_value()}."
            ),
            hint="Tag field requires ', '-separated values.",
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def invalid_reference_type_item(
        node: SDocNode,
        reference_item: Reference,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        role_and_type = (
            f"{reference_item.ref_type} / {reference_item.role}"
            if reference_item.role is not None
            else reference_item.ref_type
        )
        return StrictDocSemanticError(
            title=(
                f"Requirement relation type/role is not registered: "
                f"{role_and_type}."
            ),
            hint=f"Problematic requirement: {node.reserved_uid}.",
            example=None,
            line=node.ng_line_start,
            col=node.ng_col_start,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def invalid_marker_role(
        node: SDocNode,
        marker: Union[
            ForwardFunctionRangeMarker,
            FunctionRangeMarker,
            LineMarker,
            RangeMarker,
            ForwardRangeMarker,
        ],
        path_to_src_file: str,
    ) -> "StrictDocSemanticError":
        role_and_type = marker.role if marker.role is not None else "Any"
        return StrictDocSemanticError(
            title=f"File marker role is not registered: {role_and_type}.",
            hint=f"Problematic requirement: {node.reserved_uid}.",
            example=None,
            line=marker.ng_source_line_begin,
            col=marker.ng_source_column_begin,
            filename=path_to_src_file,
        )

    @staticmethod
    def grammar_element_has_no_mid_field(
        grammar_element: GrammarElement,
        path_to_sdoc_file: str,
    ) -> "StrictDocSemanticError":
        return StrictDocSemanticError(
            title=(
                f"Grammar element '{grammar_element.tag}' is missing the MID field "
                f"which contradicts to the DOCUMENT's ENABLE_MID setting."
            ),
            hint=(
                "Either disable the ENABLE_MID option or ensure that every element has the MID field defined."
            ),
            example=None,
            line=1,
            col=1,
            filename=path_to_sdoc_file,
        )

    @staticmethod
    def view_references_nonexisting_grammar_element(
        document: SDocDocument,
        document_view: DocumentView,
        view_element: ViewElement,
        object_type: str,
    ) -> "StrictDocSemanticError":
        assert document.meta is not None
        return StrictDocSemanticError(
            title=(
                f"View element '{view_element.view_id}' references a non-existing"
                f" grammar element '{object_type}'."
            ),
            hint=(
                "Make sure that each View element references an existing "
                "object in the grammar or the default REQUIREMENT object."
            ),
            example=None,
            line=document_view.ng_line_start,
            col=document_view.ng_col_start,
            filename=document.meta.input_doc_full_path,
        )

    @staticmethod
    def view_references_nonexisting_field(
        document: SDocDocument,
        document_view: DocumentView,
        view_element: ViewElement,
        object_type: str,
        field_name: str,
    ) -> "StrictDocSemanticError":
        assert document.meta is not None
        return StrictDocSemanticError(
            title=(
                f"View element '{view_element.view_id}' references a non-existing"
                f" field '{field_name}' for grammar element '{object_type}'."
            ),
            hint=(
                "Make sure that each View element references an existing "
                "field in the grammar for the given grammar element."
            ),
            example=None,
            line=document_view.ng_line_start,
            col=document_view.ng_col_start,
            filename=document.meta.input_doc_full_path,
        )

    @staticmethod
    def default_view_doesnt_exist(
        document: SDocDocument,
        document_config: DocumentConfig,
        default_view: str,
    ) -> "StrictDocSemanticError":
        assert document.meta is not None

        filename = document.meta.input_doc_full_path

        return StrictDocSemanticError(
            title=(
                f"Default view '{default_view}' does not exist in the document."
            ),
            hint=(
                "Make sure that the specified default view is created in the "
                "VIEWS configuration."
            ),
            example=None,
            line=document_config.ng_line_start,
            col=document_config.ng_col_start,
            filename=filename,
        )

    def to_print_message(self) -> str:
        message = ""
        message += f"error: could not parse file: {self.file_path}.\n"
        message += f"Semantic error: {self.title}\n"
        message += f"Location: {self.file_path}:{self.line}:{self.col}"
        if self.hint:
            message += f"\nHint: {self.hint}"
        if self.example:
            message += f"\nExample:\n{self.example}"
        return message



import hashlib
import io
import os
import re
import sys
from pathlib import Path
from typing import Optional, Tuple

from docutils.core import publish_parts
from docutils.parsers.rst import directives, roles
from docutils.utils import SystemMessage
from markupsafe import Markup

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.core.project_config import ProjectConfig, ProjectFeature
from strictdoc.export.rst.directives.raw_html_role import raw_html_role
from strictdoc.export.rst.directives.sphinx_style_math import (
    MathDirective,
    MathDirectiveForServer,
    eq_role,
    eq_role_for_server,
    math_role,
    math_role_for_server,
)
from strictdoc.export.rst.directives.wildcard_enhanced_image import (
    WildcardEnhancedImage,
)


class RstToHtmlFragmentWriter:
    directives.register_directive("image", WildcardEnhancedImage)

    roles.register_local_role("rawhtml", raw_html_role)

    def __init__(
        self,
        *,
        project_config: ProjectConfig,
        context_document: Optional[SDocDocument],
    ):
        self.source_path: str
        path_to_output_dir_md5: str = hashlib.md5(
            project_config.output_dir.encode("utf-8")
        ).hexdigest()

        path_to_tmp_dir = project_config.get_path_to_cache_dir()
        self.path_to_rst_cache_dir = os.path.join(
            path_to_tmp_dir, "rst", path_to_output_dir_md5
        )

        if context_document is not None:
            assert context_document.meta is not None
            WildcardEnhancedImage.current_reference_path = (
                context_document.meta.output_document_dir_full_path
            )

            # This is a delicate move. Based on a user report and our findings,
            # the csv-table RST directive relies on the 'source path' to
            # calculate paths to CSV files.
            # Our case is, however, special: we do not render RST files but
            # rather RST fragments in memory, and because of that we don't have
            # RST files to point to with 'source_path=' below.
            # At the same time, passing the output folder of the document works
            # because this RST-to-HTML writer resolves path to CSV assets
            # that are copied to that output folder by StrictDoc.
            # See CSVTable().get_csv_data() where the source_path is used.
            self.source_path = os.path.join(
                context_document.meta.output_document_dir_full_path,
                "STRICTDOC-FRAGMENT.rst",
            )
        else:
            self.source_path = "<string>"
        self.context_document: Optional[SDocDocument] = context_document

        if project_config.is_feature_activated(ProjectFeature.MATHJAX):
            if project_config.is_running_on_server:
                roles.register_canonical_role("eq", eq_role_for_server)
                roles.register_canonical_role("math", math_role_for_server)
                directives.register_directive("math", MathDirectiveForServer)
            else:
                roles.register_canonical_role("eq", eq_role)
                roles.register_canonical_role("math", math_role)
                directives.register_directive("math", MathDirective)

    def write(self, rst_fragment: str, use_cache: bool = True) -> Markup:
        assert isinstance(rst_fragment, str), rst_fragment

        # Do not try to cache very small fragments.
        if len(rst_fragment) < 40:
            return Markup(self._write_no_cache(rst_fragment))

        path_to_rst_fragment_bucket_dir = os.path.join(
            self.path_to_rst_cache_dir, str(len(rst_fragment))
        )
        fragment_md5 = hashlib.md5(rst_fragment.encode("utf-8")).hexdigest()
        path_to_cached_fragment = os.path.join(
            path_to_rst_fragment_bucket_dir, fragment_md5
        )
        if use_cache and os.path.isdir(path_to_rst_fragment_bucket_dir):
            if os.path.isfile(path_to_cached_fragment):
                with open(
                    path_to_cached_fragment, "rb"
                ) as cached_fragment_file_:
                    return Markup(cached_fragment_file_.read().decode("UTF-8"))
        else:
            Path(path_to_rst_fragment_bucket_dir).mkdir(
                parents=True, exist_ok=True
            )

        rendered_html: str = self._write_no_cache(rst_fragment)
        rendered_html_bytes = rendered_html.encode("UTF-8")

        if use_cache:
            with open(path_to_cached_fragment, "wb") as cached_fragment_file_:
                cached_fragment_file_.write(rendered_html_bytes)

        return Markup(rendered_html)

    def _write_no_cache(self, rst_fragment: str) -> str:
        assert isinstance(rst_fragment, str), rst_fragment

        # How do I convert a docutils document tree into an HTML string?
        # https://stackoverflow.com/a/32168938/598057
        # Use a io.StringIO as the warning stream to prevent warnings from
        # being printed to sys.stderr.
        # https://www.programcreek.com/python/example/88126/docutils.core.publish_parts
        warning_stream = io.StringIO()
        settings = {"warning_stream": warning_stream}

        output = publish_parts(
            rst_fragment,
            writer_name="html",
            settings_overrides=settings,
            source_path=self.source_path,
        )

        if warning_stream.tell() > 0:
            warnings = warning_stream.getvalue().rstrip("\n")
            # A typical RST warning:
            # """
            # path-to-output-folder/file.rst:4: (WARNING/2) Bullet list ends
            # without a blank line; unexpected unindent.
            # """
            match = re.search(
                r".*:(?P<line>\d+): \(.*\) (?P<message>.*)", warnings
            )
            if match is not None:
                error_message = (
                    f"RST markup syntax error on line {match.group('line')}: "
                    f"{match.group('message')}"
                )
            else:
                error_message = f"RST markup syntax error: {warnings}"
            print(  # noqa: T201
                f"error: problems when converting RST to HTML:\n{error_message}"
            )
            print("RST fragment: >>>")  # noqa: T201
            print(rst_fragment)  # noqa: T201
            print("<<<")  # noqa: T201
            sys.exit(1)

        html: str = output["html_body"]

        return html

    def write_with_validation(
        self, rst_fragment: str
    ) -> Tuple[Optional[str], Optional[str]]:
        # How do I convert a docutils document tree into an HTML string?
        # https://stackoverflow.com/a/32168938/598057
        # Use a io.StringIO as the warning stream to prevent warnings from
        # being printed to sys.stderr.
        # https://www.programcreek.com/python/example/88126/docutils.core.publish_parts
        warning_stream = io.StringIO()
        settings = {"warning_stream": warning_stream}

        try:
            output = publish_parts(
                rst_fragment, writer_name="html", settings_overrides=settings
            )
            warnings = (
                warning_stream.getvalue().rstrip("\n")
                if warning_stream.tell() > 0
                else None
            )
        except SystemMessage as exception:
            output = None
            warnings = str(exception)

        if warnings is not None and len(warnings) > 0:
            # A typical RST warning:
            # """
            # <string>:4: (WARNING/2) Bullet list ends without a blank line;
            # unexpected unindent.
            # """
            match = re.search(
                r".*<.*>:(?P<line>\d+): \(.*\) (?P<message>.*)", warnings
            )
            if match is not None:
                error_message = (
                    f"RST markup syntax error on line {match.group('line')}: "
                    f"{match.group('message')}"
                )
            else:
                error_message = f"RST markup syntax error: {warnings}"
            return None, error_message

        html = output["html_body"]

        return html, None

    @staticmethod
    def write_anchor_link(title: str, href: str) -> str:
        return f"""\
:rawhtml:`<a href="{href}">🔗&nbsp;{title}</a>`\
"""



from typing import Tuple


def get_shard(total: int, shards: int, shard: int) -> Tuple[int, int]:
    assert 0 < total, total
    assert 0 < shard <= shards, (shard, shards)
    assert shards <= total, (shards, total)

    shard_size = total // shards
    shard_size_remainder = total % shards

    left = (shard - 1) * shard_size
    right = left + shard_size
    if shard == shards:
        right += shard_size_remainder

    return left, right



# mypy: disable-error-code="unreachable"

import hashlib
import statistics
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from markupsafe import Markup

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import (
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    ParentReqReference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.git.change import (
    ChangeType,
    ChangeUnionType,
    DocumentChange,
    RequirementChange,
    RequirementFieldChange,
    SectionChange,
)
from strictdoc.helpers.cast import assert_cast, assert_optional_cast
from strictdoc.helpers.diff import get_colored_html_diff_string, similar
from strictdoc.helpers.mid import MID


def calculate_similarity(lhs: SDocNode, rhs: SDocNode) -> float:
    similar_fields = []
    for field_name_, field_values_ in lhs.ordered_fields_lookup.items():
        if field_name_ == "COMMENT":
            continue

        if field_name_ not in rhs.ordered_fields_lookup:
            continue

        if field_name_ not in ("TITLE", "STATEMENT", "RATIONALE"):
            continue

        rhs_field_values = rhs.ordered_fields_lookup[field_name_]

        lhs_field_value = field_values_[0].get_text_value()
        rhs_field_value = rhs_field_values[0].get_text_value()

        similar_fields.append(similar(lhs_field_value, rhs_field_value))

    return statistics.mean(similar_fields)


@dataclass
class ProjectTreeDiffStats:
    document_md5_hashes: Set[str] = field(default_factory=set)
    requirement_md5_hashes: Set[str] = field(default_factory=set)
    section_md5_hashes: Set[str] = field(default_factory=set)
    map_nodes_to_hashes: Dict[Any, str] = field(default_factory=dict)
    map_mid_to_nodes: Dict[MID, Any] = field(default_factory=dict)
    map_uid_to_nodes: Dict[str, Any] = field(default_factory=dict)
    map_titles_to_nodes: Dict[str, List[SDocNode]] = field(default_factory=dict)
    map_statements_to_nodes: Dict[str, Any] = field(default_factory=dict)
    map_rel_paths_to_docs: Dict[str, SDocDocument] = field(default_factory=dict)

    cache_requirement_to_requirement: Dict[SDocNode, SDocNode] = field(
        default_factory=dict
    )

    def get_md5_by_node(self, node: Any) -> str:
        return self.map_nodes_to_hashes[node]

    def contains_requirement_md5(self, requirement_md5: str) -> bool:
        return requirement_md5 in self.requirement_md5_hashes

    def contains_section_md5(self, section_md5: str) -> bool:
        return section_md5 in self.section_md5_hashes

    def contains_document_md5(self, document_md5: str) -> bool:
        return document_md5 in self.document_md5_hashes

    def get_identical_requirement_field(
        self, requirement: SDocNode, field_name: str, field_value: str
    ) -> Optional[SDocNodeField]:
        assert isinstance(field_value, str)
        other_requirement: Optional[SDocNode] = self.find_requirement(
            requirement
        )
        if other_requirement is None:
            return None

        if field_name not in other_requirement.ordered_fields_lookup:
            return None

        other_requirement_fields = other_requirement.ordered_fields_lookup[
            field_name
        ]
        for field_ in other_requirement_fields:
            if field_.get_text_value() == field_value:
                return field_
        return None

    def contains_requirement_relations(
        self,
        requirement: SDocNode,
        relation_uid: str,
        relation_role: Optional[str],
    ) -> bool:
        other_requirement: Optional[SDocNode] = self.find_requirement(
            requirement
        )
        if other_requirement is None:
            return False
        for reference_ in other_requirement.relations:
            if isinstance(reference_, (ParentReqReference, ChildReqReference)):
                if (
                    reference_.ref_uid == relation_uid
                    and reference_.role == relation_role
                ):
                    return True
        return False

    def find_requirement(self, requirement: SDocNode) -> Optional[SDocNode]:
        if requirement in self.cache_requirement_to_requirement:
            return self.cache_requirement_to_requirement[requirement]

        requirement_parent_uids = set()
        for parent_ in requirement.relations:
            if isinstance(parent_, ParentReqReference):
                requirement_parent_uids.add(parent_.ref_uid)

        if (
            requirement.mid_permanent
            and requirement.reserved_mid in self.map_mid_to_nodes
        ):
            node: SDocNode = assert_cast(
                self.map_mid_to_nodes[requirement.reserved_mid], SDocNode
            )
            return node
        elif (
            requirement.reserved_uid is None
            or requirement.reserved_uid not in self.map_uid_to_nodes
        ):
            candidate_requirements: Dict[SDocNode, float] = {}

            if (
                requirement.reserved_title is not None
                and requirement.reserved_title in self.map_titles_to_nodes
            ):
                other_requirements = self.map_titles_to_nodes[
                    requirement.reserved_title
                ]
                for other_requirement_ in other_requirements:
                    if (
                        not other_requirement_.reserved_uid
                        in requirement_parent_uids
                    ):
                        candidate_requirements[other_requirement_] = 0

            for candidate_requirement_ in candidate_requirements.keys():
                candidate_requirements[candidate_requirement_] = (
                    calculate_similarity(requirement, candidate_requirement_)
                )

            if len(candidate_requirements) > 0:
                candidate_requirement = max(
                    candidate_requirements,
                    key=lambda k: candidate_requirements[k],
                )
                self.cache_requirement_to_requirement[requirement] = (
                    candidate_requirement
                )
                return candidate_requirement
            return None

        other_requirement: SDocNode = self.map_uid_to_nodes[
            requirement.reserved_uid
        ]
        # FIXME: This is an interesting case when a SDocNode can be promoted
        # or unpromoted to a Section with the same UID preserved. Ignore this
        # case for now.
        if not isinstance(other_requirement, SDocNode):
            return None
        return other_requirement


@dataclass
class ChangeStats:
    changes: List[ChangeUnionType] = field(default_factory=list)
    _node_counter: Dict[str, Dict[ChangeType, int]] = field(
        default_factory=dict
    )
    _section_counter: Dict[ChangeType, int] = field(default_factory=dict)
    _document_counter: Dict[ChangeType, int] = field(default_factory=dict)

    map_nodes_to_changes: Dict[Any, ChangeUnionType] = field(
        default_factory=dict
    )

    def find_change(
        self, node: Any
    ) -> Optional[Union[DocumentChange, SectionChange, RequirementChange]]:
        return self.map_nodes_to_changes.get(node)

    def get_total_changes(self) -> int:
        return len(self.changes)

    def get_changes_requirements_changed(self, node_type: str) -> Optional[int]:
        all_node_changes_bucket = self._node_counter.get(node_type)
        if all_node_changes_bucket is None:
            return None
        this_node_type_changes_bucket = all_node_changes_bucket.get(
            ChangeType.REQUIREMENT
        )
        return this_node_type_changes_bucket

    def get_changes_sections_stats_string(self) -> str:
        """
        Example: 2 removed, 1 modified, 2 added.
        """
        change_components = []
        removed = self._section_counter.get(ChangeType.SECTION_REMOVED)
        if removed is not None:
            change_components.append(f"{removed} removed")
        modified = self._section_counter.get(ChangeType.SECTION_MODIFIED)
        if modified is not None:
            change_components.append(f"{modified} modified")
        added = self._section_counter.get(ChangeType.SECTION_ADDED)
        if added is not None:
            change_components.append(f"{added} added")
        assert len(change_components) > 0
        return ", ".join(change_components)

    def get_changes_requirements_stats_string(self, node_type: str) -> str:
        """
        Example: 2 removed, 1 modified, 2 added.
        """
        change_components = []

        removed_counter = self._node_counter.get(node_type)
        if removed_counter is not None:
            removed = removed_counter.get(ChangeType.REQUIREMENT_REMOVED)
            if removed is not None:
                change_components.append(f"{removed} removed")

        modified_counter = self._node_counter.get(node_type)
        if modified_counter is not None:
            modified = modified_counter.get(ChangeType.REQUIREMENT_MODIFIED)
            if modified is not None:
                change_components.append(f"{modified} modified")

        added_counter = self._node_counter.get(node_type)
        if added_counter is not None:
            added = added_counter.get(ChangeType.REQUIREMENT_ADDED)
            if added is not None:
                change_components.append(f"{added} added")
        assert len(change_components) > 0
        return ", ".join(change_components)

    def get_changes_documents_modified(self) -> Optional[int]:
        return self._document_counter.get(ChangeType.DOCUMENT)

    def get_changes_sections_modified(self) -> Optional[int]:
        return self._section_counter.get(ChangeType.SECTION)

    def add_change(
        self,
        change: Union[DocumentChange, SectionChange, RequirementChange],
        node_type: Optional[str] = None,
    ) -> None:
        self.changes.append(change)
        if change.change_type in (
            ChangeType.REQUIREMENT_REMOVED,
            ChangeType.REQUIREMENT_MODIFIED,
            ChangeType.REQUIREMENT_ADDED,
        ):
            assert isinstance(node_type, str), node_type
            node_type_counter = self._node_counter.setdefault(node_type, {})
            node_type_counter.setdefault(change.change_type, 0)
            node_type_counter[change.change_type] += 1

            node_type_counter.setdefault(ChangeType.REQUIREMENT, 0)
            node_type_counter[ChangeType.REQUIREMENT] += 1
        elif change.change_type in (
            ChangeType.SECTION_REMOVED,
            ChangeType.SECTION_MODIFIED,
            ChangeType.SECTION_ADDED,
        ):
            self._section_counter.setdefault(change.change_type, 0)
            self._section_counter[change.change_type] += 1

            self._section_counter.setdefault(ChangeType.SECTION, 0)
            self._section_counter[ChangeType.SECTION] += 1
        elif change.change_type in (ChangeType.DOCUMENT_MODIFIED,):
            self._document_counter.setdefault(change.change_type, 0)
            self._document_counter[change.change_type] += 1

            self._document_counter.setdefault(ChangeType.DOCUMENT, 0)
            self._document_counter[ChangeType.DOCUMENT] += 1

        else:  # pragma: no cover
            raise AssertionError

    @staticmethod
    def create_from_two_indexes(
        lhs_index: TraceabilityIndex,
        rhs_index: TraceabilityIndex,
        lhs_stats: ProjectTreeDiffStats,
        rhs_stats: ProjectTreeDiffStats,
    ) -> "ChangeStats":
        stats = ChangeStats()

        ChangeStats._iterate_one_index(
            lhs_index, lhs_stats, rhs_stats, stats, "left"
        )
        ChangeStats._iterate_one_index(
            rhs_index, rhs_stats, lhs_stats, stats, "right"
        )

        return stats

    @staticmethod
    def _iterate_one_index(
        index: TraceabilityIndex,
        self_stats: ProjectTreeDiffStats,
        other_stats: ProjectTreeDiffStats,
        change_stats: "ChangeStats",
        side: str,
    ) -> None:
        assert side in ("left", "right")

        for document in index.document_tree.document_list:
            #
            # The included documents are ignored. All their information should
            # be contained in the including documents at this point.
            #
            if document.document_is_included():
                continue

            #
            # First, take care of the document node itself. Check if the document
            # root-level free text (abstract) has changed.
            #
            if document not in change_stats.map_nodes_to_changes:
                other_document_or_none: Optional[SDocDocument]

                # First, the MID-based match is tried. If no MID is available,
                # try to find a document under the same path.
                if (
                    document.mid_permanent
                    and document.reserved_mid in other_stats.map_mid_to_nodes
                ):
                    other_document_or_none = other_stats.map_mid_to_nodes[
                        document.reserved_mid
                    ]
                else:
                    assert document.meta is not None
                    other_document_or_none = (
                        other_stats.map_rel_paths_to_docs.get(
                            document.meta.input_doc_rel_path.relative_path
                        )
                    )

                uid_modified: bool = False
                title_modified: bool = False
                lhs_colored_title_diff: Optional[Markup] = None
                rhs_colored_title_diff: Optional[Markup] = None

                # If there is another section and the UIDs are not the
                # same, consider the UID modified.
                # If there is no other section, consider the UID
                # modified.
                if other_document_or_none is not None:
                    if (
                        document.reserved_uid
                        != other_document_or_none.reserved_uid
                    ):
                        uid_modified = True
                else:
                    uid_modified = True

                if other_document_or_none is not None:
                    if document.title != other_document_or_none.title:
                        title_modified = True
                        lhs_colored_title_diff = get_colored_html_diff_string(
                            document.title,
                            other_document_or_none.title,
                            "left",
                        )
                        rhs_colored_title_diff = get_colored_html_diff_string(
                            document.title,
                            other_document_or_none.title,
                            "right",
                        )
                else:
                    title_modified = True

                if uid_modified or title_modified:
                    lhs_document: Optional[SDocDocument]
                    rhs_document: Optional[SDocDocument]
                    if side == "left":
                        lhs_document = document
                        rhs_document = other_document_or_none
                    else:
                        lhs_document = other_document_or_none
                        rhs_document = document

                    document_change: DocumentChange = DocumentChange(
                        matched_uid=None,
                        lhs_document=lhs_document,
                        rhs_document=rhs_document,
                        uid_modified=uid_modified,
                        title_modified=title_modified,
                        lhs_colored_title_diff=lhs_colored_title_diff,
                        rhs_colored_title_diff=rhs_colored_title_diff,
                    )
                    change_stats.map_nodes_to_changes[document] = (
                        document_change
                    )
                    if other_document_or_none is not None:
                        change_stats.map_nodes_to_changes[
                            other_document_or_none
                        ] = document_change
                    change_stats.add_change(document_change)

            document_iterator = DocumentCachingIterator(document)

            #
            # Now iterate over all nodes and collect the diff information.
            #
            # Note that document nodes appear when a document is included to
            # another document. We treat a document node as a section node
            # because they both have FREETEXT.
            #
            for node, _ in document_iterator.all_content(print_fragments=True):
                if isinstance(node, (SDocSection, SDocDocument)):
                    if node in change_stats.map_nodes_to_changes:
                        continue

                    section_md5 = self_stats.get_md5_by_node(node)
                    section_modified = not other_stats.contains_section_md5(
                        section_md5
                    )
                    if section_modified:
                        matched_mid: Optional[MID] = None
                        matched_uid: Optional[str] = None
                        other_section_or_none: Optional[SDocSection] = None

                        if (
                            node.mid_permanent
                            and node.reserved_mid
                            in other_stats.map_mid_to_nodes
                        ):
                            other_section_or_none = (
                                other_stats.map_mid_to_nodes[node.reserved_mid]
                            )
                            matched_mid = node.reserved_mid
                        if node.reserved_uid is not None:
                            assert len(node.reserved_uid) > 0
                            if other_stats.map_uid_to_nodes.get(
                                node.reserved_uid
                            ):
                                matched_uid = node.reserved_uid
                                other_section_or_none = (
                                    other_stats.map_uid_to_nodes[matched_uid]
                                )
                        # FIXME: This is when a Requirement becomes
                        # a Section with the same UID preserved.
                        if other_section_or_none is not None and not (
                            isinstance(other_section_or_none, SDocSection)
                            or (
                                isinstance(other_section_or_none, SDocNode)
                                and other_section_or_none.node_type == "SECTION"
                            )
                        ):
                            other_section_or_none = None
                            matched_uid = None
                            matched_mid = None

                        uid_modified = False
                        title_modified = False
                        lhs_colored_title_diff = None
                        rhs_colored_title_diff = None

                        # If there is another section and the UIDs are not the
                        # same, consider the UID modified.
                        # If there is no other section, consider the UID
                        # modified.
                        if other_section_or_none is not None:
                            if (
                                node.reserved_uid
                                != other_section_or_none.reserved_uid
                            ):
                                uid_modified = True
                        else:
                            uid_modified = True

                        if other_section_or_none is not None:
                            if (
                                node.reserved_title
                                != other_section_or_none.reserved_title
                            ):
                                title_modified = True
                                lhs_colored_title_diff = (
                                    get_colored_html_diff_string(
                                        node.reserved_title,
                                        other_section_or_none.reserved_title,
                                        "left",
                                    )
                                )
                                rhs_colored_title_diff = (
                                    get_colored_html_diff_string(
                                        node.reserved_title,
                                        other_section_or_none.reserved_title,
                                        "right",
                                    )
                                )
                        else:
                            title_modified = True

                        #
                        # Step: Create a section token that is used by JS to match
                        # the LHS nodes with RHS nodes.
                        #
                        section_token: Optional[str] = None
                        if other_section_or_none is not None:
                            section_token = MID.create()

                        lhs_section: Optional[
                            Union[SDocSection, SDocDocument, SDocNode]
                        ]
                        rhs_section: Optional[
                            Union[SDocSection, SDocDocument, SDocNode]
                        ]
                        if side == "left":
                            lhs_section = node
                            rhs_section = other_section_or_none
                        else:
                            lhs_section = other_section_or_none
                            rhs_section = node

                        section_change: SectionChange = SectionChange(
                            matched_mid=matched_mid,
                            matched_uid=matched_uid,
                            section_token=section_token,
                            lhs_section=lhs_section,
                            rhs_section=rhs_section,
                            uid_modified=uid_modified,
                            title_modified=title_modified,
                            lhs_colored_title_diff=lhs_colored_title_diff,
                            rhs_colored_title_diff=rhs_colored_title_diff,
                        )

                        change_stats.map_nodes_to_changes[node] = section_change
                        if other_section_or_none is not None:
                            change_stats.map_nodes_to_changes[
                                other_section_or_none
                            ] = section_change
                        change_stats.add_change(section_change)

                if isinstance(node, SDocNode):
                    #
                    # Step: We check if a requirement was modified at all, or if
                    # it has already been checked before. Skipping the requirement
                    # if there is nothing to do.
                    #
                    # FIXME: Is this 100% valid?

                    if node in change_stats.map_nodes_to_changes:
                        continue

                    requirement: SDocNode = assert_cast(node, SDocNode)
                    requirement_md5 = self_stats.get_md5_by_node(requirement)
                    requirement_modified = (
                        not other_stats.contains_requirement_md5(
                            requirement_md5
                        )
                    )
                    if not requirement_modified:
                        continue

                    other_requirement_or_none: Optional[SDocNode] = (
                        assert_optional_cast(
                            other_stats.find_requirement(requirement), SDocNode
                        )
                    )

                    requirement_change: RequirementChange

                    # If there is no other requirement to compare with,
                    # we simply record this as a trivial change where everything
                    # is tracked as "deleted" or "new".
                    if other_requirement_or_none is None:
                        if side == "left":
                            lhs_requirement = requirement
                            rhs_requirement = None
                        else:
                            lhs_requirement = None
                            rhs_requirement = requirement

                        requirement_change = RequirementChange(
                            requirement_token=None,
                            field_changes=[],
                            lhs_requirement=lhs_requirement,
                            rhs_requirement=rhs_requirement,
                        )
                        change_stats.map_nodes_to_changes[node] = (
                            requirement_change
                        )

                        change_stats.add_change(
                            requirement_change, node.node_type
                        )
                        continue

                    #
                    # Step: Starting from here, we will be looking at the
                    # difference between this and the other requirement.
                    #
                    other_requirement: SDocNode = other_requirement_or_none

                    field_changes: List[RequirementFieldChange] = []

                    #
                    # Step: Create a requirement token that is used by JS to match
                    # the LHS nodes with RHS nodes.
                    #
                    requirement_token: str = MID.create()

                    #
                    # Iterate over requirement fields.
                    #
                    requirement_fields_iter = iter(
                        requirement.enumerate_all_fields()
                    )
                    other_requirement_fields_iter = iter(
                        other_requirement.enumerate_all_fields()
                    )

                    field_checked_so_far: Set[SDocNodeField] = set()
                    while True:
                        requirement_field_tripple = next(
                            requirement_fields_iter, None
                        )
                        if (
                            requirement_field_tripple is not None
                            and requirement_field_tripple[1] != "COMMENT"
                            and requirement_field_tripple[0]
                            not in field_checked_so_far
                        ):
                            requirement_field_change = ChangeStats.create_field_change(
                                other_stats=other_stats,
                                requirement=requirement,
                                requirement_field=requirement_field_tripple[0],
                                other_requirement=other_requirement,
                                requirement_field_name=requirement_field_tripple[
                                    1
                                ],
                                requirement_field_value=requirement_field_tripple[
                                    2
                                ],
                            )
                            if requirement_field_change is not None:
                                field_changes.append(requirement_field_change)
                                if (
                                    requirement_field_change.lhs_field
                                    is not None
                                ):
                                    field_checked_so_far.add(
                                        requirement_field_change.lhs_field
                                    )
                                if (
                                    requirement_field_change.rhs_field
                                    is not None
                                ):
                                    field_checked_so_far.add(
                                        requirement_field_change.rhs_field
                                    )
                            else:
                                field_checked_so_far.add(
                                    requirement_field_tripple[0]
                                )

                        other_requirement_field_tripple = next(
                            other_requirement_fields_iter, None
                        )
                        if (
                            other_requirement_field_tripple is not None
                            and other_requirement_field_tripple[1] != "COMMENT"
                            and other_requirement_field_tripple[0]
                            not in field_checked_so_far
                        ):
                            requirement_field_change = ChangeStats.create_field_change(
                                other_stats=self_stats,
                                requirement=other_requirement,
                                requirement_field=other_requirement_field_tripple[
                                    0
                                ],
                                other_requirement=requirement,
                                requirement_field_name=other_requirement_field_tripple[
                                    1
                                ],
                                requirement_field_value=other_requirement_field_tripple[
                                    2
                                ],
                            )
                            if requirement_field_change is not None:
                                field_changes.append(requirement_field_change)
                                if (
                                    requirement_field_change.lhs_field
                                    is not None
                                ):
                                    field_checked_so_far.add(
                                        requirement_field_change.lhs_field
                                    )
                                if (
                                    requirement_field_change.rhs_field
                                    is not None
                                ):
                                    field_checked_so_far.add(
                                        requirement_field_change.rhs_field
                                    )
                            else:
                                field_checked_so_far.add(
                                    other_requirement_field_tripple[0]
                                )

                        if (
                            requirement_field_tripple is None
                            and other_requirement_field_tripple is None
                        ):
                            break

                    # COMMENT can appear in requirement several times, so
                    # it is handled separately.
                    comments_changes = ChangeStats.create_comment_field_changes(
                        requirement=requirement,
                        other_requirement=other_requirement,
                    )
                    field_changes.extend(comments_changes)

                    if side == "left":
                        lhs_requirement = requirement
                        rhs_requirement = other_requirement
                    else:
                        lhs_requirement = other_requirement
                        rhs_requirement = requirement

                    requirement_change = RequirementChange(
                        requirement_token=requirement_token,
                        field_changes=field_changes,
                        lhs_requirement=lhs_requirement,
                        rhs_requirement=rhs_requirement,
                    )

                    change_stats.map_nodes_to_changes[node] = requirement_change
                    if other_requirement_or_none is not None:
                        change_stats.map_nodes_to_changes[
                            other_requirement_or_none
                        ] = requirement_change
                    change_stats.add_change(requirement_change, node.node_type)

    @staticmethod
    def create_field_change(
        *,
        other_stats: ProjectTreeDiffStats,
        requirement: SDocNode,
        requirement_field: SDocNodeField,
        other_requirement: SDocNode,
        requirement_field_name: str,
        requirement_field_value: str,
    ) -> Optional[RequirementFieldChange]:
        assert isinstance(requirement, SDocNode)
        assert isinstance(requirement_field, SDocNodeField)
        assert isinstance(other_requirement, SDocNode)

        other_requirement_field = other_stats.get_identical_requirement_field(
            requirement, requirement_field_name, requirement_field_value
        )
        # If there is an identical field, it means the field
        # is not modified. Nothing to do.
        if other_requirement_field is not None:
            return None

        left_diff = None
        right_diff = None
        other_requirement_fields = other_requirement.ordered_fields_lookup.get(
            requirement_field_name, []
        )
        other_requirement_field = (
            other_requirement_fields[0]
            if len(other_requirement_fields) > 0
            else None
        )

        if other_requirement_field is not None:
            other_requirement_field_value = (
                other_requirement_field.get_text_value()
            )
            left_diff = get_colored_html_diff_string(
                requirement_field_value, other_requirement_field_value, "left"
            )
            right_diff = get_colored_html_diff_string(
                requirement_field_value, other_requirement_field_value, "right"
            )

        return RequirementFieldChange(
            field_name=requirement_field_name,
            lhs_field=requirement_field,
            rhs_field=other_requirement_field,
            left_diff=left_diff,
            right_diff=right_diff,
        )

    @staticmethod
    def create_comment_field_changes(
        *,
        requirement: SDocNode,
        other_requirement: SDocNode,
    ) -> List[RequirementFieldChange]:
        assert isinstance(requirement, SDocNode)
        assert isinstance(other_requirement, SDocNode)

        changes = []

        changed_fields: Dict[SDocNodeField, None] = dict.fromkeys(
            requirement.ordered_fields_lookup.get("COMMENT", []), None
        )
        changed_other_fields: Dict[SDocNodeField, None] = dict.fromkeys(
            other_requirement.ordered_fields_lookup.get("COMMENT", []), None
        )

        if len(changed_fields) == 0 or len(changed_other_fields) == 0:
            for changed_field_ in changed_fields.keys():
                changes.append(
                    RequirementFieldChange(
                        field_name="COMMENT",
                        lhs_field=changed_field_,
                        rhs_field=None,
                        left_diff=None,
                        right_diff=None,
                    )
                )
            for changed_other_field_ in changed_other_fields.keys():
                changes.append(
                    RequirementFieldChange(
                        field_name="COMMENT",
                        lhs_field=None,
                        rhs_field=changed_other_field_,
                        left_diff=None,
                        right_diff=None,
                    )
                )
            return changes

        similarities: List[Tuple[float, SDocNodeField, SDocNodeField]] = []
        for changed_field_ in list(changed_fields):
            comment_value = changed_field_.get_text_value()
            assert comment_value is not None
            for changed_other_field_ in list(changed_other_fields):
                comment_other_value = changed_other_field_.get_text_value()
                assert comment_other_value is not None

                similarity = similar(comment_value, comment_other_value)
                if similarity == 1:
                    del changed_fields[changed_field_]
                    del changed_other_fields[changed_other_field_]
                    break

                similarities.append(
                    (similarity, changed_field_, changed_other_field_)
                )

        similarities.sort(key=lambda student: student[0], reverse=True)

        for _, changed_field_, changed_other_field_ in similarities:
            if (
                changed_field_ not in changed_fields
                or changed_other_field_ not in changed_other_fields
            ):
                continue

            # This is the best change.
            comment_value = changed_field_.get_text_value()
            comment_other_value = changed_other_field_.get_text_value()
            assert comment_other_value is not None

            left_diff = get_colored_html_diff_string(
                comment_value,
                comment_other_value,
                "left",
            )
            right_diff = get_colored_html_diff_string(
                comment_value,
                comment_other_value,
                "right",
            )
            changes.append(
                RequirementFieldChange(
                    field_name="COMMENT",
                    lhs_field=changed_field_,
                    rhs_field=changed_other_field_,
                    left_diff=left_diff,
                    right_diff=right_diff,
                )
            )
            del changed_fields[changed_field_]
            del changed_other_fields[changed_other_field_]

        # Iterate over remaining fields.
        for changed_field_ in changed_fields:
            changes.append(
                RequirementFieldChange(
                    field_name="COMMENT",
                    lhs_field=changed_field_,
                    rhs_field=None,
                    left_diff=None,
                    right_diff=None,
                )
            )
        for changed_other_field_ in changed_other_fields:
            changes.append(
                RequirementFieldChange(
                    field_name="COMMENT",
                    lhs_field=None,
                    rhs_field=changed_other_field_,
                    left_diff=None,
                    right_diff=None,
                )
            )

        return changes


class ProjectDiffAnalyzer:
    @staticmethod
    def analyze_document_tree(
        traceability_index: TraceabilityIndex,
    ) -> ProjectTreeDiffStats:
        document_tree_stats: ProjectTreeDiffStats = ProjectTreeDiffStats()

        for document in traceability_index.document_tree.document_list:
            if document.document_is_included():
                continue
            ProjectDiffAnalyzer.analyze_document(document, document_tree_stats)

        return document_tree_stats

    @staticmethod
    def analyze_document(
        document: SDocDocument,
        document_tree_stats: ProjectTreeDiffStats,
    ) -> None:
        assert document.meta is not None

        document_iterator = DocumentCachingIterator(document)

        map_nodes_to_hashers: Dict[Any, Any] = {document: hashlib.md5()}

        if document.mid_permanent:
            document_tree_stats.map_mid_to_nodes[document.reserved_mid] = (
                document
            )

        document_tree_stats.map_rel_paths_to_docs[
            document.meta.input_doc_rel_path.relative_path
        ] = document

        map_nodes_to_hashers[document].update(document.title.encode("utf-8"))

        for node, _ in document_iterator.all_content(print_fragments=True):
            if node.mid_permanent:
                document_tree_stats.map_mid_to_nodes[node.reserved_mid] = node

            if isinstance(node, (SDocSection, SDocDocument)):
                if node.reserved_uid is not None:
                    document_tree_stats.map_uid_to_nodes[node.reserved_uid] = (
                        node
                    )
                hasher = hashlib.md5()
                hasher.update(node.title.encode("utf-8"))
                map_nodes_to_hashers[node] = hasher

            elif isinstance(node, SDocNode):
                if node.reserved_uid is not None:
                    document_tree_stats.map_uid_to_nodes[node.reserved_uid] = (
                        node
                    )

                hasher = hashlib.md5()
                for (
                    field_name_,
                    field_values_,
                ) in node.ordered_fields_lookup.items():
                    requirement_field: SDocNodeField = field_values_[0]
                    requirement_field_value = requirement_field.get_text_value()
                    if field_name_ == "TITLE":
                        this_title_requirements = (
                            document_tree_stats.map_titles_to_nodes.setdefault(
                                requirement_field_value, []
                            )
                        )
                        this_title_requirements.append(node)
                    elif field_name_ == "STATEMENT":
                        document_tree_stats.map_statements_to_nodes[
                            requirement_field_value
                        ] = node

                    hasher.update(requirement_field_value.encode("utf-8"))

                    # If this field appears once, there is nothing else to do.
                    if len(field_values_) == 1:
                        continue

                    # At this point, we are dealing with COMMENT because it is
                    # the only field that can appear several times.
                    for comment_field_ in field_values_[1:]:
                        hasher.update(
                            comment_field_.get_text_value().encode("utf-8")
                        )

                for reference_ in node.relations:
                    if isinstance(
                        reference_, (ParentReqReference, ChildReqReference)
                    ):
                        hasher.update(reference_.ref_uid.encode("utf-8"))
                        if reference_.role is not None:
                            hasher.update(reference_.role.encode("utf-8"))

                map_nodes_to_hashers[node] = hasher
            else:
                raise AssertionError(node)

        for node_, _ in document_iterator.all_content(print_fragments=True):
            node_md5 = map_nodes_to_hashers[node_].hexdigest().encode("utf-8")
            map_nodes_to_hashers[document].update(node_md5)

        for node_, node_hasher_ in map_nodes_to_hashers.items():
            node_md5 = node_hasher_.hexdigest()
            document_tree_stats.map_nodes_to_hashes[node_] = node_md5

            if isinstance(node_, SDocSection):
                document_tree_stats.section_md5_hashes.add(node_md5)
            elif isinstance(node_, SDocNode):
                document_tree_stats.requirement_md5_hashes.add(node_md5)
            elif isinstance(node_, SDocDocument):
                document_tree_stats.document_md5_hashes.add(node_md5)
            else:
                raise NotImplementedError(node_)



"""
@relation(SDOC-SRS-143, scope=file)
"""

from typing import Dict, Optional, Type, Union

import robot.result
from robot.api import ExecutionResult, ResultVisitor

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.reference import (
    FileEntry,
    FileEntryFormat,
    FileReference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.file_tree import File
from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.paths import path_to_posix_path


class SdocVisitor(ResultVisitor):  # type: ignore[misc]
    def __init__(self, project_config: ProjectConfig):
        self.project_config = project_config
        self.suites: Dict[str, Union[SDocDocument, SDocSection]] = {}
        self.document: Optional[SDocDocument] = None

    def visit_suite(self, suite: robot.result.TestSuite) -> None:
        """
        Create document for top level suite and sections for nested suites.
        """
        assert suite.full_name not in self.suites

        if suite.parent is None:
            self.document = SDocDocument(
                mid=None,
                title=f"Test report: {suite.name}",
                config=None,
                view=None,
                grammar=None,
                section_contents=[],
            )
            self.document.ng_including_document_reference = DocumentReference()
            grammar = DocumentGrammar.create_for_test_report(self.document)
            self.document.grammar = grammar
            self.document.config.requirement_style = "Table"
            self.suites[suite.full_name] = self.document
            summary = self.summary_from_suite(suite, self.document)
            self.document.section_contents.append(summary)
        else:
            assert self.document
            assert suite.parent.full_name in self.suites, (
                "depth-first traversal expected"
            )
            parent_sdoc_node = self.suites[suite.parent.full_name]
            section = SDocSection(
                parent=parent_sdoc_node,
                mid=None,
                uid=None,
                custom_level=None,
                title=suite.name,
                requirement_prefix=None,
                section_contents=[],
            )
            section.ng_including_document_reference = DocumentReference()
            section.ng_document_reference = DocumentReference()
            section.ng_document_reference.set_document(self.document)
            self.suites[suite.full_name] = section
            parent_sdoc_node.section_contents.append(section)
            summary_sdoc_node = self.summary_from_suite(suite, section)
            section.section_contents.append(summary_sdoc_node)

        super().visit_suite(suite)

    def visit_test(self, test: robot.result.TestCase) -> None:
        """
        Create TEST_RESULT node for each test case.
        """
        assert self.document
        assert test.parent and test.parent.full_name in self.suites, (
            "depth-first traversal expected"
        )
        parent_section = self.suites[test.parent.full_name]
        testcase_node = SDocNode(
            parent=parent_section,
            node_type="TEST_RESULT",
            fields=[],
            relations=[],
        )
        testcase_node.ng_document_reference = DocumentReference()
        testcase_node.ng_document_reference.set_document(self.document)
        testcase_node.ng_including_document_reference = DocumentReference()

        # Only executed tests will become traceable.
        if not test.skipped:
            testcase_node.set_field_value(
                field_name="UID",
                form_field_index=0,
                value=SDocNodeField(
                    parent=testcase_node,
                    field_name="UID",
                    parts=[test.full_name.upper()],
                    multiline__=None,
                ),
            )

        # Absolute path will be replaced with relative path to strictdoc root
        # in FileTraceabilityIndex.
        abs_path_on_exec_machine = path_to_posix_path(str(test.source))
        if test.source is not None:
            testcase_node.set_field_value(
                field_name="TEST_PATH",
                form_field_index=0,
                value=SDocNodeField(
                    parent=testcase_node,
                    field_name="TEST_PATH",
                    parts=[abs_path_on_exec_machine],
                    multiline__=None,
                ),
            )

        testcase_node.set_field_value(
            field_name="TEST_FUNCTION",
            form_field_index=0,
            value=SDocNodeField(
                parent=testcase_node,
                field_name="TEST_FUNCTION",
                parts=[test.name],
                multiline__=None,
            ),
        )

        testcase_node.set_field_value(
            field_name="DURATION",
            form_field_index=0,
            value=SDocNodeField(
                parent=testcase_node,
                field_name="DURATION",
                parts=[str(test.elapsed_time.total_seconds())],
                multiline__=None,
            ),
        )

        testcase_node.set_field_value(
            field_name="STATUS",
            form_field_index=0,
            value=SDocNodeField(
                parent=testcase_node,
                field_name="STATUS",
                parts=[test.status],
                multiline__=None,
            ),
        )

        testcase_node.set_field_value(
            field_name="TITLE",
            form_field_index=0,
            value=SDocNodeField(
                parent=testcase_node,
                field_name="TITLE",
                parts=[test.name],
                multiline__=None,
            ),
        )

        if not test.skipped:
            # File path will be resolved in FileTraceabilityIndex.
            testcase_node.relations.append(
                FileReference(
                    parent=testcase_node,
                    g_file_entry=FileEntry(
                        parent=None,
                        g_file_format=FileEntryFormat.SOURCECODE,
                        g_file_path="#FORWARD#",
                        g_line_range=None,
                        function=test.name,
                        clazz=None,
                    ),
                )
            )

        parent_section.section_contents.append(testcase_node)

    def summary_from_suite(
        self,
        suite: robot.result.TestSuite,
        parent: Union[SDocDocument, SDocSection],
    ) -> SDocNode:
        assert self.document
        summary_table = f"""\
.. list-table:: Test suite summary
    :widths: 25 10
    :header-rows: 0

    * - **Number of tests:**
      - {suite.statistics.total}
    * - **Number of successful tests:**
      - {suite.statistics.passed}
    * - **Number of failed tests:**
      - {suite.statistics.failed}
    * - **Number of skipped tests:**
      - {suite.statistics.skipped}
"""
        summary_node = SDocNode(
            parent=parent,
            node_type="TEXT",
            fields=[],
            relations=[],
        )
        summary_node.ng_document_reference = DocumentReference()
        summary_node.ng_document_reference.set_document(self.document)
        summary_node.ng_including_document_reference = DocumentReference()
        summary_node.set_field_value(
            field_name="STATEMENT",
            form_field_index=0,
            value=SDocNodeField(
                parent=summary_node,
                field_name="STATEMENT",
                parts=[summary_table],
                multiline__="True",
            ),
        )
        return summary_node


class RobotOutputXMLReader:
    @classmethod
    def read_from_file(
        cls: Type["RobotOutputXMLReader"],
        doc_file: File,
        project_config: ProjectConfig,
    ) -> SDocDocument:
        execution_result = ExecutionResult(doc_file.full_path)
        sdoc_visitor = SdocVisitor(project_config)
        execution_result.visit(sdoc_visitor)
        if sdoc_visitor.document is None:
            raise RuntimeError(
                f"No test suite could be parsed from {doc_file.full_path}"
            )
        return sdoc_visitor.document



"""
@relation(SDOC-SRS-28, scope=file)
"""

import datetime
from copy import deepcopy
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    RelationMarkerType,
    SourceFileTraceabilityInfo,
)
from strictdoc.core.asset_manager import AssetManager
from strictdoc.core.constants import GraphLinkType
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.file_dependency_manager import FileDependencyManager
from strictdoc.core.file_traceability_index import FileTraceabilityIndex
from strictdoc.core.graph.abstract_bucket import ALL_EDGES
from strictdoc.core.graph_database import GraphDatabase
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile
from strictdoc.core.transforms.validation_error import (
    SingleValidationError,
)
from strictdoc.core.tree_cycle_detector import TreeCycleDetector
from strictdoc.helpers.cast import assert_cast, assert_optional_cast
from strictdoc.helpers.file_modification_time import set_file_modification_time
from strictdoc.helpers.mid import MID
from strictdoc.helpers.ordered_set import OrderedSet
from strictdoc.helpers.paths import SDocRelativePath
from strictdoc.helpers.sorting import alphanumeric_sort


class TraceabilityIndex:
    def __init__(
        self,
        document_tree: DocumentTree,
        document_iterators: Dict[SDocDocument, DocumentCachingIterator],
        file_traceability_index: FileTraceabilityIndex,
        graph_database: GraphDatabase,
        file_dependency_manager: FileDependencyManager,
    ):
        self._document_iterators: Dict[
            SDocDocument, DocumentCachingIterator
        ] = document_iterators
        self._file_traceability_index: FileTraceabilityIndex = (
            file_traceability_index
        )

        self.graph_database: GraphDatabase = graph_database
        self.document_tree: DocumentTree = document_tree
        self.asset_manager: Optional[AssetManager] = None
        self.file_dependency_manager: FileDependencyManager = (
            file_dependency_manager
        )
        self.index_last_updated = datetime.datetime.today()
        self.contains_included_documents = False
        self.strictdoc_last_update: datetime.datetime = (
            datetime.datetime.fromtimestamp(0)
        )

        # The timestamp is used by HTML/JS for invalidating the search index
        # cache in the IndexedDB database.
        # If no documents have to be re-generated with the second+ run of
        # StrictDoc, this timestamp is set of a modification date of the first
        # SDoc document, see export_static_html_search_index.
        self.search_index_timestamp: datetime.datetime = datetime.datetime.now(
            datetime.timezone.utc
        )

    @property
    def document_iterators(self) -> Dict[SDocDocument, DocumentCachingIterator]:
        return self._document_iterators

    def is_small_project(self) -> bool:
        """
        Check if project is small to control feature availability.

        This method helps to decide if StrictDoc will precompile Jinja templates
        to Python files or not. Precompilation may take half a second time, so
        it is only worth doing it when a project is relatively large.

        Below, making some assumptions about what makes a small or larger
        project.
        """
        if len(self.document_tree.document_list) >= 3:
            return False
        for document_ in self.document_tree.document_list:
            if len(document_.section_contents) > 5:
                return False
        return (
            self.graph_database.get_count(
                link_type=GraphLinkType.NODE_TO_PARENT_NODES
            )
            <= 10
        )

    def has_parent_requirements(self, requirement: SDocNode) -> bool:
        assert isinstance(requirement, SDocNode)
        if not isinstance(requirement.reserved_uid, str):
            return False

        parent_requirements = self.graph_database.get_link_values(
            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
            lhs_node=requirement,
            edge=ALL_EDGES,
        )
        return len(parent_requirements) > 0

    def has_children_requirements(self, requirement: SDocNode) -> bool:
        assert isinstance(requirement, SDocNode)
        if not isinstance(requirement.reserved_uid, str):
            return False

        children_requirements = self.graph_database.get_link_values(
            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
            lhs_node=requirement,
            edge=ALL_EDGES,
        )
        return len(children_requirements) > 0

    def has_source_file_reqs(self, source_file_rel_path: str) -> bool:
        return self._file_traceability_index.has_source_file_reqs(
            source_file_rel_path
        )

    def has_node_connections(self, node_uid: str) -> bool:
        assert isinstance(node_uid, str), node_uid
        return self.graph_database.has_link(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=node_uid,
        )

    def get_node_by_mid(self, node_mid: MID) -> Any:
        assert isinstance(node_mid, MID), node_mid
        return self.graph_database.get_link_value(
            link_type=GraphLinkType.MID_TO_NODE, lhs_node=node_mid
        )

    def get_node_by_mid_weak(self, node_mid: MID) -> Optional[Any]:
        assert isinstance(node_mid, MID), node_mid
        return self.graph_database.get_link_value_weak(
            link_type=GraphLinkType.MID_TO_NODE, lhs_node=node_mid
        )

    def get_file_traceability_index(self) -> FileTraceabilityIndex:
        return self._file_traceability_index

    def get_document_iterator(
        self, document: SDocDocument
    ) -> DocumentCachingIterator:
        return self.document_iterators[document]

    def get_parent_requirements(self, requirement: SDocNode) -> List[SDocNode]:
        assert isinstance(requirement, SDocNode)
        if not isinstance(requirement.reserved_uid, str):
            return []

        return list(
            self.graph_database.get_link_values(
                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                lhs_node=requirement,
                edge=ALL_EDGES,
            )
        )

    def get_parent_relations_with_roles(
        self, node: SDocNode
    ) -> List[Tuple[SDocNode, Optional[str]]]:
        assert isinstance(node, SDocNode)

        return list(
            self.graph_database.get_link_values_with_edges(
                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                lhs_node=node,
                edge=ALL_EDGES,
            )
        )

    def get_parent_relations_with_role(
        self, requirement: SDocNode, role: Optional[str]
    ) -> List[Tuple[SDocNode, Optional[str]]]:
        assert isinstance(requirement, SDocNode)
        if requirement.reserved_uid is None:
            return []

        return list(
            self.graph_database.get_link_values_with_edges(
                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                lhs_node=requirement,
                edge=role,
            )
        )

    def get_child_relations_with_roles(
        self, requirement: SDocNode
    ) -> List[Tuple[SDocNode, Optional[str]]]:
        assert isinstance(requirement, SDocNode)

        return list(
            self.graph_database.get_link_values_with_edges(
                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                lhs_node=requirement,
                edge=ALL_EDGES,
            )
        )

    def get_child_relations_with_role(
        self, requirement: SDocNode, role: Optional[str]
    ) -> List[Tuple[SDocNode, Optional[str]]]:
        assert isinstance(requirement, SDocNode)
        if requirement.reserved_uid is None:
            return []

        return list(
            self.graph_database.get_link_values_with_edges(
                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                lhs_node=requirement,
                edge=role,
            )
        )

    def get_children_requirements(
        self, requirement: SDocNode
    ) -> List[SDocNode]:
        assert isinstance(requirement, SDocNode)
        if not isinstance(requirement.reserved_uid, str):
            return []

        return list(
            self.graph_database.get_link_values(
                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                lhs_node=requirement,
                edge=ALL_EDGES,
            )
        )

    def has_tags(self, document: SDocDocument) -> bool:
        return self.graph_database.has_link(
            link_type=GraphLinkType.DOCUMENT_TO_TAGS,
            lhs_node=document.reserved_mid,
        )

    def get_counted_tags(
        self, document: SDocDocument
    ) -> Generator[Tuple[str, int], None, None]:
        document_tags_or_none = self.graph_database.get_link_value(
            link_type=GraphLinkType.DOCUMENT_TO_TAGS,
            lhs_node=document.reserved_mid,
        )
        document_tags: Dict[str, int] = assert_cast(document_tags_or_none, dict)

        tags = sorted(document_tags.keys(), key=alphanumeric_sort)
        for tag in tags:
            yield tag, document_tags[tag]

    def get_requirement_file_links(
        self, requirement: SDocNode
    ) -> List[Tuple[str, List[RelationMarkerType]]]:
        return self._file_traceability_index.get_requirement_file_links(
            requirement
        )

    def get_source_file_reqs(
        self, source_file_rel_path: str
    ) -> Optional[List[SDocNode]]:
        return self._file_traceability_index.get_source_file_reqs(
            source_file_rel_path
        )

    def get_coverage_info(
        self, source_file_rel_path: str
    ) -> SourceFileTraceabilityInfo:
        return self._file_traceability_index.get_coverage_info(
            source_file_rel_path
        )

    def get_coverage_info_weak(
        self, source_file_rel_path: str
    ) -> Optional[SourceFileTraceabilityInfo]:
        return self._file_traceability_index.get_coverage_info_weak(
            source_file_rel_path
        )

    def get_node_by_uid(self, uid: str) -> Any:
        assert isinstance(uid, str) and len(uid) > 0, uid
        return self.graph_database.get_link_value(
            link_type=GraphLinkType.UID_TO_NODE, lhs_node=uid
        )

    def get_node_by_uid_weak2(self, uid: str) -> Optional[Any]:
        """
        FIXME: This can likely replace _weak below with no problem.
        """
        assert isinstance(uid, str) and len(uid) > 0, uid
        return self.graph_database.get_link_value_weak(
            link_type=GraphLinkType.UID_TO_NODE, lhs_node=uid
        )

    def get_linkable_node_by_uid(
        self, uid: str
    ) -> Union[SDocDocument, SDocNode, SDocSection, Anchor]:
        return assert_cast(
            self.get_node_by_uid(uid),
            (SDocDocument, SDocNode, SDocSection, Anchor),
        )

    def get_node_by_uid_weak(
        self, uid: str
    ) -> Union[SDocDocument, SDocSection, SDocNode, None]:
        assert isinstance(uid, str), uid
        for document in self.document_tree.document_list:
            document_iterator = DocumentCachingIterator(document)
            for node, _ in document_iterator.all_content(print_fragments=False):
                if isinstance(node, SDocDocument):
                    if node.config.uid == uid:
                        return node
                elif isinstance(node, SDocSection):
                    if node.reserved_uid == uid:
                        return node
                elif isinstance(node, SDocNode):
                    if node.reserved_uid == uid:
                        return node
                else:
                    raise NotImplementedError
        return None

    def get_linkable_node_by_uid_weak(
        self, uid: str
    ) -> Union[SDocDocument, SDocNode, SDocSection, Anchor, None]:
        return assert_optional_cast(
            self.graph_database.get_link_value_weak(
                link_type=GraphLinkType.UID_TO_NODE, lhs_node=uid
            ),
            (SDocDocument, SDocNode, SDocSection, Anchor),
        )

    def get_incoming_links(
        self, node: Union[SDocDocument, SDocNode, SDocSection, Anchor]
    ) -> Optional[List[InlineLink]]:
        incoming_links = self.graph_database.get_link_values(
            link_type=GraphLinkType.NODE_TO_INCOMING_LINKS,
            lhs_node=node.reserved_mid,
        )
        if incoming_links is None or len(incoming_links) == 0:
            return None
        # FIXME: Should the graph database return OrderedSet or a copied list()?
        return list(incoming_links)

    def create_traceability_info(
        self,
        source_file: SourceFile,
        traceability_info: SourceFileTraceabilityInfo,
    ) -> None:
        self._file_traceability_index.create_traceability_info(
            source_file, traceability_info
        )

    def create_document(self, document: SDocDocument) -> None:
        assert isinstance(document, SDocDocument)
        if document.reserved_uid is not None:
            self.graph_database.create_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=document.reserved_uid,
                rhs_node=document,
            )
        self.graph_database.create_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=document.reserved_mid,
            rhs_node=document,
        )

    def create_section(self, section: SDocSection) -> None:
        assert isinstance(section, SDocSection)
        if section.reserved_uid is not None:
            self.graph_database.create_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=section.reserved_uid,
                rhs_node=section,
            )
        self.graph_database.create_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=section.reserved_mid,
            rhs_node=section,
        )

    def create_inline_link(self, new_link: InlineLink) -> None:
        assert isinstance(new_link, InlineLink)

        # InlineLink points to a section, node or to anchor.
        assert self.graph_database.has_link(
            link_type=GraphLinkType.UID_TO_NODE, lhs_node=new_link.link
        )

        node_or_anchor: Union[SDocDocument, SDocNode, SDocSection, Anchor] = (
            assert_cast(
                self.graph_database.get_link_value(
                    link_type=GraphLinkType.UID_TO_NODE,
                    lhs_node=new_link.link,
                ),
                (SDocDocument, SDocNode, SDocSection, Anchor),
            )
        )
        self.graph_database.create_link(
            link_type=GraphLinkType.NODE_TO_INCOMING_LINKS,
            lhs_node=node_or_anchor.reserved_mid,
            rhs_node=new_link,
        )
        self.graph_database.create_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=new_link.reserved_mid,
            rhs_node=new_link,
        )

    def update_last_updated(self) -> None:
        """
        Update the index's last updated date to the current time.

        This is a rather broad way of signalling that all documents of the index
        need to be re-generated when they are opened next time. Several UI
        actions use this method to ensure a complete re-generation of all
        documents.
        """
        self.index_last_updated = datetime.datetime.today()

    def create_requirement(self, requirement: SDocNode) -> None:
        assert isinstance(requirement, SDocNode), requirement

        self.graph_database.create_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=requirement.reserved_mid,
            rhs_node=requirement,
        )
        if requirement.reserved_uid is not None:
            self.graph_database.create_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=requirement.reserved_uid,
                rhs_node=requirement,
            )

    def update_requirement_uid(
        self, requirement: SDocNode, old_uid: Optional[str]
    ) -> None:
        if old_uid is None:
            if requirement.reserved_uid:
                self.graph_database.create_link(
                    link_type=GraphLinkType.UID_TO_NODE,
                    lhs_node=requirement.reserved_uid,
                    rhs_node=requirement,
                )
            return

        self.graph_database.delete_link(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=old_uid,
            rhs_node=requirement,
        )

        if requirement.reserved_uid is not None:
            self.graph_database.create_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=requirement.reserved_uid,
                rhs_node=requirement,
            )

    def update_requirement_parent_uid(
        self, requirement: SDocNode, parent_uid: str, role: Optional[str]
    ) -> None:
        assert requirement.reserved_uid is not None
        assert isinstance(parent_uid, str), parent_uid
        assert role is None or len(role) > 0, role
        parent_nodes: OrderedSet[SDocNode] = (
            self.graph_database.get_link_values(
                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                lhs_node=requirement,
                edge=role,
            )
        )
        # If a relation to the parent uid through a given role already exists,
        # there is nothing to do.
        if any(node_.reserved_uid == parent_uid for node_ in parent_nodes):
            return

        parent_requirement: SDocNode = self.graph_database.get_link_value(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=parent_uid,
        )

        document = assert_cast(requirement.get_document(), SDocDocument)
        parent_requirement_document = assert_cast(
            parent_requirement.get_document(), SDocDocument
        )

        self.graph_database.create_link(
            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
            lhs_node=requirement,
            rhs_node=parent_requirement,
            edge=role,
        )
        self.graph_database.create_link(
            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
            lhs_node=parent_requirement,
            rhs_node=requirement,
            edge=role,
        )

        cycle_detector = TreeCycleDetector()

        def parent_cycle_traverse_(node_id: str) -> List[Any]:
            node = self.graph_database.get_link_value(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=node_id,
            )
            return list(
                map(
                    lambda node_: node_.reserved_uid,
                    self.graph_database.get_link_values(
                        link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                        lhs_node=node,
                    ),
                )
            )

        cycle_detector.check_node(
            requirement.reserved_uid,
            parent_cycle_traverse_,
        )

        # Mark document and parent document (if different) for re-generation.
        assert document.meta is not None
        set_file_modification_time(
            document.meta.input_doc_full_path, datetime.datetime.today()
        )
        if parent_requirement_document != document:
            assert parent_requirement_document.meta is not None
            set_file_modification_time(
                parent_requirement_document.meta.input_doc_full_path,
                datetime.datetime.today(),
            )

    def update_requirement_child_uid(
        self, requirement: SDocNode, child_uid: str, role: Optional[str]
    ) -> None:
        assert requirement.reserved_uid is not None
        assert isinstance(child_uid, str), child_uid
        assert role is None or len(role) > 0, role
        child_nodes: OrderedSet[SDocNode] = self.graph_database.get_link_values(
            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
            lhs_node=requirement,
            edge=role,
        )

        # If a relation to the parent uid through a given role already exists,
        # there is nothing to do.
        if any(node_.reserved_uid == child_uid for node_ in child_nodes):
            return
        child_requirement = self.graph_database.get_link_value(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=child_uid,
        )

        document = assert_cast(requirement.get_document(), SDocDocument)
        child_requirement_document = assert_cast(
            child_requirement.get_document(), SDocDocument
        )

        self.graph_database.create_link(
            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
            lhs_node=child_requirement,
            rhs_node=requirement,
            edge=role,
        )
        self.graph_database.create_link(
            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
            lhs_node=requirement,
            rhs_node=child_requirement,
            edge=role,
        )

        cycle_detector = TreeCycleDetector()

        def child_cycle_traverse_(node_id: str) -> List[Any]:
            node = self.graph_database.get_link_value(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=node_id,
            )
            return list(
                map(
                    lambda node_: node_.reserved_uid,
                    self.graph_database.get_link_values(
                        link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                        lhs_node=node,
                    ),
                )
            )

        cycle_detector.check_node(
            requirement.reserved_uid,
            child_cycle_traverse_,
        )

        # Mark document and parent document (if different) for re-generation.
        assert document.meta is not None
        set_file_modification_time(
            document.meta.input_doc_full_path, datetime.datetime.today()
        )
        if child_requirement_document != document:
            assert child_requirement_document.meta is not None
            set_file_modification_time(
                child_requirement_document.meta.input_doc_full_path,
                datetime.datetime.today(),
            )

    def update_with_anchor(self, anchor: Anchor) -> None:
        # By this time, we know that the validations have passed just before.
        existing_anchor: Optional[Anchor] = (
            self.graph_database.get_link_value_weak(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=anchor.value,
            )
        )
        if existing_anchor is not None:
            self.graph_database.delete_link(
                link_type=GraphLinkType.MID_TO_NODE,
                lhs_node=existing_anchor.mid,
                rhs_node=existing_anchor,
            )
            self.graph_database.delete_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=existing_anchor.value,
                rhs_node=existing_anchor,
            )

        self.graph_database.create_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=anchor.mid,
            rhs_node=anchor,
        )
        self.graph_database.create_link(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=anchor.value,
            rhs_node=anchor,
        )

    def update_disconnect_two_documents_if_no_links_left(
        self, document: SDocDocument, other_document: SDocDocument
    ) -> None:
        assert document != other_document

        for node, _ in self.document_iterators[document].all_content(
            print_fragments=False
        ):
            if not isinstance(node, SDocNode):
                continue

            requirement_node: SDocNode = node

            # If a requirement has no UID, it cannot contribute to any relation-based
            # connection between any two documents.
            if requirement_node.reserved_uid is None:
                continue

            requirement_parents: OrderedSet[SDocNode] = (
                self.graph_database.get_link_values(
                    link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                    lhs_node=requirement_node,
                )
            )

            # If at least one parent or child relation points to the other
            # document, terminate, not deleting the link between documents.
            for parent_requirement in requirement_parents:
                parent_requirement_document: SDocDocument = assert_cast(
                    parent_requirement.get_document(), SDocDocument
                )
                if parent_requirement_document == other_document:
                    return

            requirement_children = self.graph_database.get_link_values(
                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                lhs_node=requirement_node,
            )

            for child_requirement_, _ in requirement_children:
                if child_requirement_.document == other_document:
                    return

    def delete_document(self, document: SDocDocument) -> None:
        assert isinstance(document, SDocDocument), document

        self.graph_database.delete_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=document.reserved_mid,
            rhs_node=document,
        )
        if document.reserved_uid is not None:
            self.graph_database.delete_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=document.reserved_uid,
                rhs_node=document,
            )

    def delete_section(self, section: SDocSection) -> None:
        assert isinstance(section, SDocSection), section

        self.graph_database.delete_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=section.reserved_mid,
            rhs_node=section,
        )
        if section.reserved_uid is not None:
            self.graph_database.delete_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=section.reserved_uid,
                rhs_node=section,
            )

    def delete_requirement(self, requirement: SDocNode) -> None:
        assert isinstance(requirement, SDocNode), SDocNode

        self.graph_database.delete_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=requirement.reserved_mid,
            rhs_node=requirement,
        )
        if requirement.reserved_uid is not None:
            self.graph_database.delete_link(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=requirement.reserved_uid,
                rhs_node=requirement,
            )
            self.graph_database.delete_all_links(
                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                lhs_node=requirement,
            )
            self.graph_database.delete_all_links(
                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                lhs_node=requirement,
            )

    def remove_requirement_parent_uid(
        self, requirement: SDocNode, parent_uid: str, role: Optional[str]
    ) -> None:
        assert requirement.reserved_uid is not None
        assert isinstance(parent_uid, str), parent_uid
        assert role is None or len(role) > 0, role

        parent_requirement: SDocNode = self.graph_database.get_link_value(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=parent_uid,
        )

        self.graph_database.delete_link(
            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
            lhs_node=requirement,
            rhs_node=parent_requirement,
            edge=role,
        )
        self.graph_database.delete_link(
            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
            lhs_node=parent_requirement,
            rhs_node=requirement,
            edge=role,
        )

        document = assert_cast(requirement.get_document(), SDocDocument)
        parent_requirement_document = assert_cast(
            parent_requirement.get_document(), SDocDocument
        )

        # If there are no requirements linking between the documents,
        # remove the link.
        if document != parent_requirement_document:
            self.update_disconnect_two_documents_if_no_links_left(
                document, parent_requirement_document
            )

        # Mark document and parent document (if different) for re-generation.
        assert document.meta is not None
        set_file_modification_time(
            document.meta.input_doc_full_path, datetime.datetime.today()
        )
        if parent_requirement_document != document:
            assert parent_requirement_document.meta is not None
            set_file_modification_time(
                parent_requirement_document.meta.input_doc_full_path,
                datetime.datetime.today(),
            )

    def remove_requirement_child_uid(
        self, requirement: SDocNode, child_uid: str, role: Optional[str]
    ) -> None:
        assert requirement.reserved_uid is not None
        assert isinstance(child_uid, str), child_uid
        assert role is None or len(role) > 0, role

        child_requirement: SDocNode = self.graph_database.get_link_value(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=child_uid,
        )
        document: SDocDocument = assert_cast(
            requirement.get_document(), SDocDocument
        )
        child_requirement_document: SDocDocument = assert_cast(
            child_requirement.get_document(), SDocDocument
        )

        self.graph_database.delete_link(
            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
            lhs_node=requirement,
            rhs_node=child_requirement,
            edge=role,
        )
        self.graph_database.delete_link(
            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
            lhs_node=child_requirement,
            rhs_node=requirement,
            edge=role,
        )

        # If there are no requirements linking between the documents,
        # remove the link.
        if document != child_requirement_document:
            self.update_disconnect_two_documents_if_no_links_left(
                document, child_requirement_document
            )

        # Mark document and parent document (if different) for re-generation.
        assert document.meta is not None
        set_file_modification_time(
            document.meta.input_doc_full_path, datetime.datetime.today()
        )
        if child_requirement_document != document:
            assert child_requirement_document.meta is not None
            set_file_modification_time(
                child_requirement_document.meta.input_doc_full_path,
                datetime.datetime.today(),
            )

    def remove_inline_link(self, inline_link: InlineLink) -> None:
        sections_with_incoming_links = (
            self.graph_database.get_link_values_reverse(
                link_type=GraphLinkType.NODE_TO_INCOMING_LINKS,
                rhs_node=inline_link,
            )
        )

        for node_with_incoming_links in list(sections_with_incoming_links):
            self.graph_database.delete_link(
                link_type=GraphLinkType.NODE_TO_INCOMING_LINKS,
                lhs_node=node_with_incoming_links,
                rhs_node=inline_link,
            )

        self.graph_database.delete_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=inline_link.reserved_mid,
            rhs_node=inline_link,
        )

    def remove_anchor_by_uid(self, anchor_uid: str) -> None:
        anchor: Anchor = self.graph_database.get_link_value(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=anchor_uid,
        )
        self.graph_database.delete_link(
            link_type=GraphLinkType.MID_TO_NODE,
            lhs_node=anchor.mid,
            rhs_node=anchor,
        )
        self.graph_database.delete_link(
            link_type=GraphLinkType.UID_TO_NODE,
            lhs_node=anchor_uid,
            rhs_node=anchor,
        )

    def validate_node_against_anchors(
        self,
        *,
        node: Union[SDocDocument, SDocSection, SDocNode, None],
        new_anchors: List[Anchor],
    ) -> None:
        assert node is None or isinstance(
            node, (SDocDocument, SDocSection, SDocNode)
        )
        assert isinstance(new_anchors, list)

        # Check that this node does not have duplicated anchors.
        new_anchor_uids = set()
        for anchor in new_anchors:
            if anchor.value in new_anchor_uids:
                raise SingleValidationError(
                    "A node cannot have two anchors with "
                    f"the same identifier: {anchor.value}."
                )
            new_anchor_uids.add(anchor.value)

        # If the node is new, the validation is easier: we just need to make
        # sure that there are no existing anchors with the UIDs brought
        # by the new anchors.
        if node is None:
            for anchor_uid in new_anchor_uids:
                if self.graph_database.has_link(
                    link_type=GraphLinkType.UID_TO_NODE, lhs_node=anchor_uid
                ):
                    raise SingleValidationError(
                        "A node contains an anchor that already exists: "
                        f"{anchor_uid}."
                    )
            return

        # If the node is an existing node, we need to check that:
        # 1) If some of the new anchors already exist in the project tree, we
        #    need to ensure that they exist in the current node, otherwise we
        #    raise a duplication validation error.
        # 2) If the new anchors do not contain some of the existing node's
        #    current anchors, this means these anchors are being removed. In
        #    that case, we need to check if these anchors are used by any LINKs,
        #    raising a validation if they do.
        existing_node_anchor_uids = set()

        # FIXME: No test reaches this for Section or Document.
        assert isinstance(node, SDocNode)
        for node_anchor_ in node.get_anchors():
            existing_node_anchor_uids.add(node_anchor_.value)

        #
        # Validation 1: Assert all UIDs are either:
        # a) new
        # b) exist in this node
        # c) raise a duplication error.
        #
        for anchor_uid in new_anchor_uids:
            if (
                self.graph_database.has_link(
                    link_type=GraphLinkType.UID_TO_NODE, lhs_node=anchor_uid
                )
                and anchor_uid not in existing_node_anchor_uids
            ):
                duplicate_anchor: Anchor = self.graph_database.get_link_value(
                    link_type=GraphLinkType.UID_TO_NODE, lhs_node=anchor_uid
                )
                node_with_duplicate_anchor: SDocNode = assert_cast(
                    duplicate_anchor.parent_node(), SDocNode
                )
                raise SingleValidationError(
                    "Another node contains an anchor with the same UID: "
                    f"{anchor_uid}. {node_with_duplicate_anchor.get_display_node_type()}: "
                    f"{node_with_duplicate_anchor.get_display_title()}."
                )

        #
        # Validation 2: Check that removed anchors do not have any incoming
        # links.
        #
        to_be_removed_anchor_uids = existing_node_anchor_uids - new_anchor_uids
        for to_be_removed_anchor_uid_ in to_be_removed_anchor_uids:
            to_be_removed_anchor: Anchor = self.graph_database.get_link_value(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=to_be_removed_anchor_uid_,
            )
            incoming_links = self.graph_database.get_link_values(
                link_type=GraphLinkType.NODE_TO_INCOMING_LINKS,
                lhs_node=to_be_removed_anchor.reserved_mid,
            )
            if incoming_links is not None and len(incoming_links) > 0:
                incoming_link: InlineLink = incoming_links[0]
                incoming_link_parent_node = incoming_link.parent_node()
                raise SingleValidationError(
                    f"Cannot remove anchor with UID "
                    f"'{incoming_link.link}' because it has incoming "
                    f"links. Containing node: "
                    f"{incoming_link_parent_node.get_display_node_type()}: "
                    f"'{incoming_link_parent_node.get_display_title()}'."
                )

    def validate_can_remove_node(
        self, *, node: Union[SDocNode, Anchor]
    ) -> None:
        incoming_links: Optional[List[InlineLink]] = self.get_incoming_links(
            node
        )
        if incoming_links is not None and len(incoming_links) > 0:
            link_list_message = ", ".join(
                map(
                    lambda l_: f"'{l_.parent_node().get_display_title()}' -> '{l_.link}'",
                    incoming_links,
                )
            )
            raise SingleValidationError(
                f"Cannot remove node '{node.get_display_title()}' with incoming LINKs from: {link_list_message}."
            )
        if isinstance(node, SDocNode):
            child_nodes: List[SDocNode] = self.get_children_requirements(node)
            if child_nodes is not None and len(child_nodes) > 0:
                nodes_list_message = ", ".join(
                    map(
                        lambda n_: "'" + n_.get_display_title() + "'",
                        child_nodes,
                    )
                )
                raise SingleValidationError(
                    f"Cannot remove node '{node.get_display_title()}' "
                    f"with incoming relations from:\n{nodes_list_message}."
                )

    # FIXME: This function will be removed in 2025-Q3.
    def validate_section_can_remove_uid(
        self, *, section: SDocSection
    ) -> None:  # pragma: no cover
        section_incoming_links: Optional[List[InlineLink]] = (
            self.get_incoming_links(section)
        )
        if section_incoming_links is None or len(section_incoming_links) == 0:
            return

        unique_incoming_link_parent_nodes = []
        for section_incoming_link in section_incoming_links:
            unique_incoming_link_parent_nodes.append(
                section_incoming_link.parent.parent
            )
        incoming_link_parent_titles = map(
            lambda n: f"'{n.title}'", unique_incoming_link_parent_nodes
        )
        incoming_links_sources_string = ", ".join(incoming_link_parent_titles)
        raise SingleValidationError(
            f"Cannot remove a section UID "
            f"'{section.reserved_uid}' because there are "
            f"existing LINKs referencing this "
            "section. The incoming links are located in these "
            f"nodes: {incoming_links_sources_string}."
        )

    def validate_can_create_uid(
        self, uid: str, existing_node_mid: Optional[MID]
    ) -> None:
        assert isinstance(uid, str), uid
        assert len(uid) > 0, uid

        existing_node_with_uid: Union[
            SDocDocument, SDocSection, SDocNode, None
        ] = self.get_node_by_uid_weak(uid)

        if existing_node_with_uid is None:
            return

        if existing_node_mid is not None:
            if existing_node_with_uid.reserved_mid == existing_node_mid:
                return

        raise SingleValidationError(
            "UID uniqueness validation error: "
            "There is already an existing node "
            f"with this UID: {existing_node_with_uid.get_display_title()}."
        )

    def clone_to_bundle_document(
        self, project_config: ProjectConfig
    ) -> Tuple["TraceabilityIndex", SDocDocument]:
        """
        Clone this traceability index and create a new bundle document.

        The only use case for this method is the generation of a bundle document.
        Since the bundle document does not exist on file system, some parameters
        are set artificially:
        - The bundle is assumed to be an input file in the root input folder.
        - The bundle is generated to the root of the output folder (level=0).
        - Some variables do not contribute (yet) to the final result, so they
          are marked as NOT_RELEVANT.
        """
        traceability_index_copy = deepcopy(self)
        bundle_document = SDocDocument(
            mid=None,
            title=project_config.project_title,
            config=None,
            view=None,
            grammar=None,
            section_contents=[],
            is_bundle_document=True,
        )
        bundle_document.config = DocumentConfig.default_config(bundle_document)

        if (
            project_config.bundle_document_version is not None
            and len(project_config.bundle_document_version) > 0
        ):
            bundle_document.config.version = (
                project_config.bundle_document_version
            )

        if (
            project_config.bundle_document_date is not None
            and len(project_config.bundle_document_date) > 0
        ):
            bundle_document.config.date = project_config.bundle_document_date

        bundle_document.meta = DocumentMeta(
            level=0,
            file_tree_mount_folder="NOT_RELEVANT",
            document_filename="bundle.sdoc",
            document_filename_base="bundle",
            input_doc_full_path="NOT_RELEVANT",
            input_doc_rel_path=SDocRelativePath("bundle.sdoc"),
            input_doc_dir_rel_path=SDocRelativePath(""),
            input_doc_assets_dir_rel_path=SDocRelativePath("NOT_RELEVANT"),
            output_document_dir_full_path=project_config.export_output_html_root,
            output_document_dir_rel_path=SDocRelativePath(""),
        )
        traceability_index_copy.document_iterators[bundle_document] = (
            DocumentCachingIterator(bundle_document)
        )
        for document_ in traceability_index_copy.document_tree.document_list:
            # Ignore all included documents. They are anyway included by
            # the including documents.
            if document_.document_is_included():
                continue

            assert document_.ng_including_document_reference is not None
            document_.ng_including_document_reference.set_document(
                bundle_document
            )
            bundle_document.section_contents.append(document_)
        traceability_index_copy.document_tree.document_list = [bundle_document]
        bundle_document.ng_including_document_reference = DocumentReference()
        return traceability_index_copy, bundle_document



"""
@relation(SDOC-SRS-35, scope=file)
"""

from typing import Optional

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.html_templates import HTMLTemplates, JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer


class SourceCoverageViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
    ) -> None:
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config
        self.link_renderer: LinkRenderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )
        self.standalone: bool = False
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__

    def get_document_level(self) -> int:
        return 0

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/source_file_coverage/index.jinja", view_object=self
        )

    def render_static_url(self, url: str) -> str:
        return self.link_renderer.render_static_url(url)

    def render_url(self, url: str) -> str:
        # FIXME: Switch to Markup(...).
        return self.link_renderer.render_url(url)

    def get_file_stats_lines_total(self, source_file: SourceFile) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return "0"
        return str(trace_info.file_stats.lines_total)

    def get_file_stats_lines_total_non_empty(
        self, source_file: SourceFile
    ) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return "0"
        return str(trace_info.file_stats.lines_non_empty)

    def get_file_stats_non_empty_lines_covered(
        self, source_file: SourceFile
    ) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return "0"
        return str(trace_info.ng_lines_covered)

    def get_file_stats_non_empty_lines_covered_percentage(
        self, source_file: SourceFile
    ) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return f"{0:.1f}"
        covered = trace_info.ng_lines_covered
        total_non_empty = trace_info.file_stats.lines_non_empty
        percentage = (
            (covered / total_non_empty * 100) if total_non_empty > 0 else 0
        )
        return f"{percentage:.1f}"

    def get_file_stats_functions_total(self, source_file: SourceFile) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return "0"
        return str(len(trace_info.functions))

    def get_file_stats_functions_covered(self, source_file: SourceFile) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return "0"
        return str(trace_info.covered_functions)

    def get_file_stats_functions_covered_percentage(
        self, source_file: SourceFile
    ) -> str:
        trace_info: Optional[SourceFileTraceabilityInfo] = (
            self.traceability_index.get_coverage_info_weak(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )
        if trace_info is None:
            return f"{0:.1f}"
        covered = trace_info.covered_functions
        total = len(trace_info.functions)
        percentage = (covered / total * 100) if total > 0 else 0
        return f"{percentage:.1f}"


class SourceFileCoverageHTMLGenerator:
    @staticmethod
    def export(
        *,
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        html_templates: HTMLTemplates,
    ) -> Markup:
        view_object = SourceCoverageViewObject(
            traceability_index=traceability_index,
            project_config=project_config,
        )
        return view_object.render_screen(html_templates.jinja_environment())



from collections import Counter
from typing import Any, List


def find_duplicates(lst: List[Any]) -> List[Any]:
    counter = Counter(lst)
    return [item for item, count in counter.items() if count > 1]



import os
from pathlib import Path
from typing import List, Union

from strictdoc.backend.excel.excel_import import ExcelImport
from strictdoc.backend.reqif.reqif_import import ReqIFImport
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.writer import SDWriter
from strictdoc.cli.cli_arg_parser import (
    ImportExcelCommandConfig,
    ImportReqIFCommandConfig,
)
from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.string import (
    create_safe_document_file_name,
)
from strictdoc.helpers.timing import timing_decorator


class ImportAction:
    @staticmethod
    @timing_decorator("Import")
    def do_import(
        import_config: Union[
            ImportReqIFCommandConfig, ImportExcelCommandConfig
        ],
        project_config: ProjectConfig,
    ) -> None:
        if isinstance(import_config, ImportReqIFCommandConfig):
            reqif_documents: List[SDocDocument] = ReqIFImport.import_from_file(
                import_config, project_config
            )
            for document_ in reqif_documents:
                Path(import_config.output_path).mkdir(
                    parents=True, exist_ok=True
                )
                path_to_output_document = os.path.join(
                    import_config.output_path,
                    create_safe_document_file_name(document_.reserved_title)
                    + ".sdoc",
                )
                document_content = SDWriter(project_config).write(document_)
                with open(
                    path_to_output_document, "w", encoding="utf8"
                ) as output_file:
                    output_file.write(document_content)
        elif isinstance(import_config, ImportExcelCommandConfig):
            excel_document: SDocDocument = ExcelImport.import_from_file(
                import_config
            )
            Path(import_config.output_path).mkdir(parents=True, exist_ok=True)
            path_to_output_document = os.path.join(
                import_config.output_path,
                os.path.splitext(os.path.basename(import_config.input_path))[0]
                + ".sdoc",
            )
            document_content = SDWriter(project_config).write(excel_document)
            with open(
                path_to_output_document, "w", encoding="utf8"
            ) as output_file:
                output_file.write(document_content)
        else:
            raise NotImplementedError()  # pragma: no cover



import sys
import traceback
from typing import Tuple


def get_exception_origin() -> Tuple[str, int, str, str]:
    """
    Get the last traceback entry where the exception occurred.
    """

    tb = traceback.extract_tb(sys.exc_info()[2])[-1]
    filename, lineno, func, text = tb
    return filename, lineno, func, text


class ExceptionInfo:
    def __init__(self, exception: Exception):
        self.exception = exception

        filename, lineno, func, _ = get_exception_origin()
        self.filename = filename
        self.lineno = lineno
        self.func = func

        exc_type, exc_value, exc_traceback = sys.exc_info()

        traceback_strings = traceback.format_exception(
            exc_type, exc_value, exc_traceback, limit=None, chain=True
        )

        self.traceback_str = "".join(traceback_strings)

    def get_stack_trace(self) -> str:
        return self.traceback_str

    def get_detailed_error_message(self) -> str:
        return (
            f"error: {str(self.exception)}\n"
            f"error source: {self.filename}:{self.lineno}, function: {self.func}()"
        )


class StrictDocException(Exception):
    def __str__(self) -> str:
        return self.to_print_message()

    def to_print_message(self) -> str:
        return str(self.args[0])


class StrictDocChildProcessException(Exception):
    """
    A special custom exception that wraps the ExceptionInfo.

    RATIONALE: This is needed to pass exception information from the child
    process to the parent process. The reason is that ProcessPoolExecutor does
    not propagate the exception stack trace to the parent process. StrictDoc
    captures the child process's trace as ExceptionInfo to address this limitation.
    """

    def __init__(self, exception_info: ExceptionInfo):
        self.exception_info: ExceptionInfo = exception_info



"""
@relation(SDOC-SRS-26, scope=file)
"""

from collections import OrderedDict
from typing import Any, Generator, List, Optional, Tuple, Union

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    ReferenceType,
)
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.model import (
    RESERVED_NON_META_FIELDS,
    RequirementFieldName,
    SDocDocumentIF,
    SDocElementIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    ParentReqReference,
    Reference,
)
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.mid import MID
from strictdoc.helpers.string import ensure_newline


@auto_described
class SDocNodeContext:
    def __init__(self) -> None:
        self.title_number_string: Optional[str] = None


@auto_described
class SDocNodeField:
    def __init__(
        self,
        parent: Optional["SDocNode"],
        field_name: str,
        parts: List[Any],
        multiline__: Optional[str],
    ) -> None:
        self.parent: Optional[SDocNode] = parent
        self.field_name: str = field_name
        self.parts: List[Any] = parts
        self.multiline: bool = multiline__ is not None and len(multiline__) > 0

    @staticmethod
    def create_from_string(
        parent: Optional["SDocNode"],
        field_name: str,
        field_value: str,
        multiline: bool,
    ) -> "SDocNodeField":
        assert isinstance(field_name, str) and len(field_name) > 0, field_name
        assert isinstance(field_value, str) and len(field_value) > 0, (
            field_value
        )

        return SDocNodeField(
            parent=parent,
            field_name=field_name,
            parts=[field_value],
            multiline__="multiline" if multiline else None,
        )

    def is_multiline(self) -> bool:
        return self.multiline

    def get_text_value(self) -> str:
        text = ""
        for part in self.parts:
            if isinstance(part, str):
                text += part
            elif isinstance(part, InlineLink):
                text += "[LINK: "
                text += part.link
                text += "]"
            elif isinstance(part, Anchor):
                text += "[ANCHOR: "
                text += part.value
                if part.has_title:
                    text += ", "
                    text += part.title
                text += "]"
                text += "\n"
            else:
                raise NotImplementedError(part)  # pragma: no cover
        return text


@auto_described
class SDocNode(SDocNodeIF):
    def __init__(
        self,
        parent: Union[SDocDocumentIF, SDocSectionIF, SDocNodeIF],
        node_type: str,
        fields: List[SDocNodeField],
        relations: List[Reference],
        is_composite: bool = False,
        section_contents: Optional[List[SDocElementIF]] = None,
        node_type_close: Optional[str] = None,
    ) -> None:
        assert parent
        assert isinstance(node_type, str)
        assert isinstance(relations, list), relations

        self.parent: Union[SDocDocumentIF, SDocSectionIF, SDocNodeIF] = parent

        self.node_type: str = node_type

        if node_type_close is not None and len(node_type_close) > 0:
            if node_type != node_type_close:
                raise StrictDocException(
                    "[[NODE]] syntax error: "
                    "Opening and closing tags must match: "
                    f"opening: {node_type}, closing: {node_type_close}."
                )
            assert is_composite
        else:
            assert not is_composite

        self.is_composite: bool = is_composite

        ordered_fields_lookup: OrderedDict[str, List[SDocNodeField]] = (
            OrderedDict()
        )

        has_meta: bool = False
        for field in fields:
            if field.field_name not in RESERVED_NON_META_FIELDS:
                has_meta = True
            ordered_fields_lookup.setdefault(field.field_name, []).append(field)

        self.section_contents: List[SDocElementIF] = (
            section_contents if section_contents is not None else []
        )

        self.relations: List[Reference] = relations

        # TODO: Is it worth to move this to dedicated Presenter* classes to
        # keep this class textx-only?
        self.has_meta: bool = has_meta

        # This property is only used for validating fields against grammar
        # during TextX parsing and processing.
        self.fields_as_parsed = fields

        self.ordered_fields_lookup: OrderedDict[str, List[SDocNodeField]] = (
            ordered_fields_lookup
        )
        self.ng_level: Optional[int] = None
        self.ng_document_reference: Optional[DocumentReference] = None
        self.ng_including_document_reference: Optional[DocumentReference] = None
        self.ng_line_start: Optional[int] = None
        self.ng_line_end: Optional[int] = None
        self.ng_col_start: Optional[int] = None
        self.ng_col_end: Optional[int] = None
        self.ng_byte_start: Optional[int] = None
        self.ng_byte_end: Optional[int] = None
        self.context: SDocNodeContext = SDocNodeContext()

        mid: Optional[str] = None
        mid_fields: Optional[List[SDocNodeField]] = ordered_fields_lookup.get(
            "MID", None
        )
        if mid_fields is not None:
            mid = mid_fields[0].get_text_value()
        self.reserved_mid: MID = MID(mid) if mid is not None else MID.create()
        self.mid_permanent: bool = mid is not None

        self.ng_resolved_custom_level: Optional[str] = None
        self.custom_level: Optional[str] = None
        if RequirementFieldName.LEVEL in ordered_fields_lookup:
            level = ordered_fields_lookup[RequirementFieldName.LEVEL][
                0
            ].get_text_value()
            self.ng_resolved_custom_level = level
            self.custom_level = level

        # This is always true, unless the node is filtered out with --filter-requirements.
        self.ng_whitelisted: bool = True

        self.ng_has_requirements: bool = False

    @staticmethod
    def get_type_string() -> str:
        return "requirement"

    def get_node_type_string(self) -> Optional[str]:
        return self.node_type

    def get_display_node_type_string(self) -> Optional[str]:
        if self.is_composite:
            return f"[[{self.node_type}]]"
        return f"[{self.node_type}]"

    def get_display_title(
        self,
        include_toc_number: bool = True,  # noqa: ARG002
    ) -> str:
        if self.reserved_title is not None:
            return self.reserved_title
        if self.reserved_uid is not None:
            return self.reserved_uid
        if self.node_type == "TEXT":
            if isinstance(self.parent, SDocSectionIF):
                return f'Text node from section "{self.parent.get_display_title()}"'
            if isinstance(self.parent, SDocDocumentIF):
                return f'Text node from document "{self.parent.get_display_title()}"'
        return f"{self.node_type} with no title/UID"

    @property
    def is_root_included_document(self) -> bool:
        return False

    @property
    def is_root(self) -> bool:
        document = assert_cast(self.get_document(), SDocDocumentIF)
        return document.config.root is True

    def has_multiline_fields(self) -> bool:
        """
        FIXME: It should be possible to avoid calculating this every time.
        """

        document = assert_cast(self.get_document(), SDocDocumentIF)
        grammar = assert_cast(document.grammar, DocumentGrammar)
        element: GrammarElement = grammar.elements_by_type[self.node_type]

        for fields_ in self.ordered_fields_lookup.values():
            for field_ in fields_:
                if element.is_field_multiline(field_.field_name):
                    return True
        return False

    def has_any_text_nodes(self) -> bool:
        # The workaround: hasattr(...) makes mypy happy.
        return any(
            node_.__class__.__name__ == "SDocNode"
            and hasattr(node_, "node_type")
            and node_.node_type == "TEXT"
            for node_ in self.section_contents
        )

    #
    # Reserved fields
    #

    @property
    def reserved_uid(self) -> Optional[str]:
        return self._get_cached_field(
            RequirementFieldName.UID, singleline_only=True
        )

    @reserved_uid.setter
    def reserved_uid(self, uid: Optional[str]) -> None:
        self.set_field_value(
            field_name="UID",
            form_field_index=0,
            value=uid,
        )

    @property
    def reserved_status(self) -> Optional[str]:
        return self._get_cached_field(
            RequirementFieldName.STATUS, singleline_only=True
        )

    @property
    def reserved_tags(self) -> Optional[List[str]]:
        if RequirementFieldName.TAGS not in self.ordered_fields_lookup:
            return None
        field: SDocNodeField = self.ordered_fields_lookup[
            RequirementFieldName.TAGS
        ][0]
        assert not field.is_multiline(), (
            f"Field {RequirementFieldName.TAGS} must be a single-line field."
        )
        tags = field.get_text_value().split(", ")
        return tags

    @property
    def reserved_title(self) -> Optional[str]:
        return self._get_cached_field(
            RequirementFieldName.TITLE, singleline_only=True
        )

    def has_reserved_statement(self) -> bool:
        document = assert_cast(self.get_document(), SDocDocumentIF)
        grammar = assert_cast(document.grammar, DocumentGrammar)
        element: GrammarElement = grammar.elements_by_type[self.node_type]
        return element.content_field[0] in self.ordered_fields_lookup

    @property
    def reserved_statement(self) -> Optional[str]:
        document = assert_cast(self.get_document(), SDocDocumentIF)
        grammar = assert_cast(document.grammar, DocumentGrammar)
        element: GrammarElement = grammar.elements_by_type[self.node_type]
        return self._get_cached_field(
            element.content_field[0], singleline_only=False
        )

    @property
    def rationale(self) -> Optional[str]:
        return self._get_cached_field(
            RequirementFieldName.RATIONALE, singleline_only=False
        )

    def is_requirement(self) -> bool:
        return True

    def is_normative_node(self) -> bool:
        return not self.is_text_node()

    def is_text_node(self) -> bool:
        return self.node_type == "TEXT"

    def is_section(self) -> bool:
        return False

    def is_document(self) -> bool:
        return False

    def get_document(self) -> Optional[SDocDocumentIF]:
        assert self.ng_document_reference is not None, self
        return self.ng_document_reference.get_document()

    def get_including_document(self) -> Optional[SDocDocumentIF]:
        assert self.ng_including_document_reference is not None
        return self.ng_including_document_reference.get_document()

    def get_parent_or_including_document(self) -> SDocDocumentIF:
        assert self.ng_including_document_reference is not None
        including_document_or_none = (
            self.ng_including_document_reference.get_document()
        )
        if including_document_or_none is not None:
            return including_document_or_none

        assert self.ng_document_reference is not None
        document: Optional[SDocDocumentIF] = (
            self.ng_document_reference.get_document()
        )
        assert document is not None, (
            "A valid requirement must always have a reference to the document."
        )
        return document

    def get_display_node_type(self) -> str:
        return "Node"

    def get_debug_info(self) -> str:
        debug_components: List[str] = []
        if self.reserved_mid is not None:
            debug_components.append(f"MID = '{self.reserved_mid}'")
        if self.reserved_uid is not None:
            debug_components.append(f"UID = '{self.reserved_uid}'")
        if self.reserved_title is not None:
            debug_components.append(f"TITLE = '{self.reserved_title}'")

        document: Optional[SDocDocumentIF] = self.get_document()
        if document is not None:
            debug_components.append(f"document = {document.get_debug_info()}")
        return f"Requirement({', '.join(debug_components)})"

    def document_is_included(self) -> bool:
        assert self.ng_including_document_reference is not None
        return self.ng_including_document_reference.get_document() is not None

    def get_requirement_style_mode(self) -> str:
        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        grammar = assert_cast(document.grammar, DocumentGrammar)
        element: GrammarElement = grammar.elements_by_type[self.node_type]
        if node_style := element.get_view_style():
            return node_style
        return document.config.get_requirement_style_mode()

    def get_content_field_name(self) -> str:
        document = assert_cast(self.get_document(), SDocDocumentIF)
        grammar = assert_cast(document.grammar, DocumentGrammar)

        element: GrammarElement = grammar.elements_by_type[self.node_type]
        return element.content_field[0]

    def get_content_field(self) -> SDocNodeField:
        document = assert_cast(self.get_document(), SDocDocumentIF)
        grammar = assert_cast(document.grammar, DocumentGrammar)

        element: GrammarElement = grammar.elements_by_type[self.node_type]
        return self.ordered_fields_lookup[element.content_field[0]][0]

    def get_field_by_name(self, field_name: str) -> SDocNodeField:
        return self.ordered_fields_lookup[field_name][0]

    def get_anchors(self) -> List[Anchor]:
        this_node_anchors: List[Anchor] = []
        for field_ in self.enumerate_fields():
            for field_part_ in field_.parts:
                if isinstance(field_part_, Anchor):
                    this_node_anchors.append(field_part_)
        return this_node_anchors

    def get_comment_fields(self) -> List[SDocNodeField]:
        if RequirementFieldName.COMMENT not in self.ordered_fields_lookup:
            return []
        return self.ordered_fields_lookup[RequirementFieldName.COMMENT]

    def get_requirement_references(self, ref_type: str) -> List[Reference]:
        if len(self.relations) == 0:
            return []
        references: List[Reference] = []
        for reference in self.relations:
            if reference.ref_type != ref_type:
                continue
            references.append(reference)
        return references

    def get_requirement_reference_uids(
        self,
    ) -> List[Tuple[str, str, Optional[str]]]:
        if len(self.relations) == 0:
            return []
        references: List[Tuple[str, str, Optional[str]]] = []
        for reference in self.relations:
            if reference.ref_type == ReferenceType.PARENT:
                parent_reference: ParentReqReference = assert_cast(
                    reference, ParentReqReference
                )
                references.append(
                    (
                        parent_reference.ref_type,
                        parent_reference.ref_uid,
                        parent_reference.role,
                    )
                )
            elif reference.ref_type == ReferenceType.CHILD:
                child_reference: ChildReqReference = assert_cast(
                    reference, ChildReqReference
                )
                references.append(
                    (
                        child_reference.ref_type,
                        child_reference.ref_uid,
                        child_reference.role,
                    )
                )
        return references

    def enumerate_fields(self) -> Generator[SDocNodeField, None, None]:
        requirement_fields = self.ordered_fields_lookup.values()
        for requirement_field_list in requirement_fields:
            yield from requirement_field_list

    def enumerate_all_fields(
        self,
    ) -> Generator[Tuple[SDocNodeField, str, str], None, None]:
        for field in self.enumerate_fields():
            meta_field_value = field.get_text_value()
            yield field, field.field_name, meta_field_value

    def enumerate_meta_fields(
        self, skip_single_lines: bool = False, skip_multi_lines: bool = False
    ) -> Generator[Tuple[str, SDocNodeField], None, None]:
        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )

        document_grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )

        element: GrammarElement = document_grammar.elements_by_type[
            self.node_type
        ]

        for field in self.enumerate_fields():
            if field.field_name in RESERVED_NON_META_FIELDS:
                continue

            is_single_line_field = not element.is_field_multiline(
                field.field_name
            )

            if is_single_line_field and skip_single_lines:
                continue
            if (not is_single_line_field) and skip_multi_lines:
                continue

            field_human_title = element.fields_map[field.field_name]
            yield field_human_title.get_field_human_name(), field

    def get_meta_field_value_by_title(self, field_title: str) -> Optional[str]:
        assert isinstance(field_title, str)
        if field_title not in self.ordered_fields_lookup:
            return None
        field: SDocNodeField = self.ordered_fields_lookup[field_title][0]
        return field.get_text_value()

    def get_field_human_title(self, field_name: str) -> str:
        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        document_grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )
        element: GrammarElement = document_grammar.elements_by_type[
            self.node_type
        ]
        field_human_title = element.fields_map[field_name]
        return field_human_title.get_field_human_name()

    def get_field_human_title_for_statement(self) -> str:
        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )
        element: GrammarElement = grammar.elements_by_type[self.node_type]
        field_human_title = element.fields_map[element.content_field[0]]
        return field_human_title.get_field_human_name()

    def get_prefix(self) -> Optional[str]:
        if (
            own_prefix := self._get_cached_field(
                RequirementFieldName.PREFIX, singleline_only=True
            )
        ) is not None:
            if own_prefix == "None":
                return None
            return own_prefix

        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )
        element: GrammarElement = grammar.elements_by_type[self.node_type]
        if (element_prefix := element.property_prefix) is not None:
            if element_prefix == "None":
                return None
            return element_prefix

        # FIXME: Is this a reasonable behavior?
        if (
            isinstance(self.parent, SDocNode)
            and self.parent.node_type == "SECTION"
        ):
            if (parent_prefix := self.parent.get_prefix()) is not None:
                return parent_prefix
            return document.get_prefix()

        return self.parent.get_prefix()

    def get_prefix_for_new_node(self, node_type: str) -> Optional[str]:
        assert isinstance(node_type, str) and len(node_type), node_type

        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )
        element: GrammarElement = grammar.elements_by_type[node_type]
        if (element_prefix := element.property_prefix) is not None:
            if element_prefix == "None":
                return None
            return element_prefix

        return self.get_prefix()

    def dump_fields_as_parsed(self) -> str:
        # FIXME:
        # - The name of the method can be improved (used in error messages).
        # - fields can diverge from fields_as_parsed.
        return ", ".join(
            list(
                map(
                    lambda r: r.field_name,
                    self.fields_as_parsed,
                )
            )
        )

    def blacklist_if_needed(self) -> None:
        if self.section_contents is not None:
            for node_ in self.section_contents:
                if node_.ng_whitelisted:
                    return

        self.ng_whitelisted = False

        # If it turns out that all child nodes are blacklisted,
        # go up and blacklist the parent node if needed.
        if self.parent.ng_whitelisted:
            self.parent.blacklist_if_needed()

    def _get_cached_field(
        self, field_name: str, singleline_only: bool
    ) -> Optional[str]:
        if field_name not in self.ordered_fields_lookup:
            return None
        field: SDocNodeField = self.ordered_fields_lookup[field_name][0]

        if singleline_only and field.is_multiline():
            raise NotImplementedError(
                f"Field {field_name} must be a single-line field."
            )

        return field.get_text_value()

    # Below all mutating methods.

    def set_field_value(
        self,
        *,
        field_name: str,
        form_field_index: int,
        value: Optional[Union[str, SDocNodeField]],
    ) -> None:
        """
        Create or update a field by name with the given value.

        The purpose of this purpose is to provide a single-method API for
        updating any field of a requirement. A requirement might use only some
        fields of a document grammar, so an extra exercise done by the method is
        to ensure that an added field that has not been attached to the
        requirement before will be put at the right index.
        """
        assert isinstance(field_name, str)

        # If a field value is being removed, there is not much to do.
        if value is None or (isinstance(value, str) and len(value) == 0):
            # Comment is a special because there can be multiple comments.
            # Empty comments are simply ignored and do not show up in the
            # updated requirement.
            if field_name == RequirementFieldName.COMMENT:
                return

            if field_name in self.ordered_fields_lookup:
                del self.ordered_fields_lookup[field_name]
            return

        # If a field value is being added or updated.
        document: SDocDocumentIF = assert_cast(
            self.get_document(), SDocDocumentIF
        )
        grammar: DocumentGrammar = assert_cast(
            document.grammar, DocumentGrammar
        )
        element: GrammarElement = grammar.elements_by_type[self.node_type]

        field_index = element.field_titles.index(field_name)

        multiline = element.is_field_multiline(field_name)
        if multiline and isinstance(value, str):
            value = ensure_newline(value)

        if field_name in self.ordered_fields_lookup:
            if len(self.ordered_fields_lookup[field_name]) > form_field_index:
                self.ordered_fields_lookup[field_name][form_field_index] = (
                    SDocNodeField.create_from_string(
                        self,
                        field_name=field_name,
                        field_value=value,
                        multiline=multiline,
                    )
                    if isinstance(value, str)
                    else value
                )
            else:
                self.ordered_fields_lookup[field_name].insert(
                    form_field_index,
                    SDocNodeField.create_from_string(
                        self,
                        field_name=field_name,
                        field_value=value,
                        multiline=multiline,
                    )
                    if isinstance(value, str)
                    else value,
                )
            return

        new_ordered_fields_lookup = OrderedDict()
        for field_title in element.field_titles[:field_index]:
            if field_title in self.ordered_fields_lookup:
                new_ordered_fields_lookup[field_title] = (
                    self.ordered_fields_lookup[field_title]
                )
        new_ordered_fields_lookup[field_name] = [
            SDocNodeField.create_from_string(
                self,
                field_name=field_name,
                field_value=value,
                multiline=multiline,
            )
            if isinstance(value, str)
            else value
        ]
        after_field_index = field_index + 1
        for field_title in element.field_titles[after_field_index:]:
            if field_title in self.ordered_fields_lookup:
                new_ordered_fields_lookup[field_title] = (
                    self.ordered_fields_lookup[field_title]
                )
        self.ordered_fields_lookup = new_ordered_fields_lookup
        self._update_has_meta()

    def _update_has_meta(self) -> None:
        has_meta: bool = False
        for field in self.enumerate_fields():
            if field.field_name not in RESERVED_NON_META_FIELDS:
                has_meta = True
        self.has_meta = has_meta


@auto_described
class SDocCompositeNode(SDocNode):
    def __init__(
        self,
        parent: Union[SDocDocumentIF, SDocSectionIF, SDocNodeIF],
        **fields: Any,
    ) -> None:
        super().__init__(parent, **fields, is_composite=True)



from markupsafe import Markup

from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.view_objects.project_tree_view_object import (
    ProjectTreeViewObject,
)
from strictdoc.export.html.html_templates import HTMLTemplates


class ProjectMapGenerator:
    @staticmethod
    def export(
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        html_templates: HTMLTemplates,
    ) -> Markup:
        assert isinstance(html_templates, HTMLTemplates)

        view_object = ProjectTreeViewObject(
            traceability_index=traceability_index,
            project_config=project_config,
        )
        return view_object.render_map(html_templates.jinja_environment())



"""
@relation(SDOC-SRS-51, scope=file)
"""

import os.path
from subprocess import CompletedProcess, TimeoutExpired, run
from typing import List, Tuple

from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.timing import measure_performance


class PDFPrintDriver:
    @staticmethod
    def get_pdf_from_html(
        project_config: ProjectConfig,
        paths_to_print: List[Tuple[str, str]],
    ) -> None:
        assert isinstance(paths_to_print, list), paths_to_print
        path_to_html2pdf4doc_cache = os.path.join(
            project_config.get_path_to_cache_dir(), "html2pdf"
        )
        cmd: List[str] = [
            # Using sys.executable instead of "python" is important because
            # venv subprocess call to python resolves to wrong interpreter,
            # https://github.com/python/cpython/issues/86207
            # Switching back to calling html2pdf4doc directly because the
            # python -m doesn't work well with PyInstaller.
            # sys.executable, "-m"
            "html2pdf4doc",
            "print",
            "--cache-dir",
            path_to_html2pdf4doc_cache,
        ]
        if project_config.chromedriver is not None:
            cmd.extend(
                [
                    "--chromedriver",
                    project_config.chromedriver,
                ]
            )
        for path_to_print_ in paths_to_print:
            cmd.append(path_to_print_[0])
            cmd.append(path_to_print_[1])

        with measure_performance(
            "PDFPrintDriver: printing HTML to PDF using HTML2PDF and Chrome Driver"
        ):
            try:
                _: CompletedProcess[bytes] = run(
                    cmd,
                    capture_output=False,
                    check=False,
                )
            except TimeoutExpired:
                raise TimeoutError from None



"""
@relation(SDOC-SRS-72, scope=file)
"""

import os.path
from typing import List

from reqif.parser import ReqIFParser, ReqIFZParser
from reqif.reqif_bundle import ReqIFBundle, ReqIFZBundle

from strictdoc.backend.reqif.p01_sdoc.reqif_to_sdoc_converter import (
    P01_ReqIFToSDocConverter,
)
from strictdoc.backend.reqif.sdoc_reqif_fields import ReqIFProfile
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.cli.cli_arg_parser import ImportReqIFCommandConfig
from strictdoc.core.project_config import ProjectConfig


class ReqIFImport:
    @staticmethod
    def import_from_file(
        import_config: ImportReqIFCommandConfig, project_config: ProjectConfig
    ) -> List[SDocDocument]:
        converter = ReqIFImport.select_reqif_profile(import_config)

        assert os.path.isfile(import_config.input_path), (
            import_config.input_path
        )

        documents: List[SDocDocument]
        if import_config.input_path.endswith(".reqifz"):
            reqifz_bundle: ReqIFZBundle = ReqIFZParser.parse(
                import_config.input_path
            )
            assert len(reqifz_bundle.reqif_bundles) > 0
            documents = converter.convert_reqif_bundle(
                next(iter(reqifz_bundle.reqif_bundles.values())),
                enable_mid=import_config.reqif_enable_mid
                or project_config.reqif_enable_mid,
                import_markup=import_config.reqif_import_markup
                if import_config.reqif_import_markup is not None
                else project_config.reqif_import_markup,
            )
        else:
            reqif_bundle: ReqIFBundle = ReqIFParser.parse(
                import_config.input_path
            )
            documents = converter.convert_reqif_bundle(
                reqif_bundle,
                enable_mid=import_config.reqif_enable_mid
                or project_config.reqif_enable_mid,
                import_markup=import_config.reqif_import_markup
                if import_config.reqif_import_markup is not None
                else project_config.reqif_import_markup,
            )
        return documents

    @staticmethod
    def select_reqif_profile(
        import_config: ImportReqIFCommandConfig,
    ) -> P01_ReqIFToSDocConverter:
        if (
            import_config.profile is None
            or import_config.profile == ReqIFProfile.P01_SDOC
        ):
            return P01_ReqIFToSDocConverter()
        raise NotImplementedError(
            f"Unsupported profile: {import_config.profile}"
        )



import re
from collections import defaultdict
from typing import Dict, List, Optional

from markupsafe import Markup
from starlette.datastructures import FormData

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.form_data import ParsedFormData, parse_form_data
from strictdoc.helpers.string import sanitize_html_form_field
from strictdoc.server.error_object import ErrorObject
from strictdoc.server.helpers.turbo import render_turbo_stream


@auto_described
class IncludedDocumentFormObject(ErrorObject):
    def __init__(
        self,
        *,
        document_mid: str,
        context_document_mid: str,
        document_title: str,
        jinja_environment: JinjaEnvironment,
    ) -> None:
        assert isinstance(document_mid, str), document_mid
        assert isinstance(context_document_mid, str), context_document_mid
        assert isinstance(document_title, str), document_title
        super().__init__()
        self.document_mid: Optional[str] = document_mid
        self.context_document_mid: Optional[str] = context_document_mid
        self.document_title: str = document_title
        self.jinja_environment: JinjaEnvironment = jinja_environment

    @staticmethod
    def create_from_request(
        *, request_form_data: FormData, jinja_environment: JinjaEnvironment
    ) -> "IncludedDocumentFormObject":
        request_form_data_as_list = [
            (field_name, field_value)
            for field_name, field_value in request_form_data.multi_items()
        ]
        request_form_dict: ParsedFormData = assert_cast(
            parse_form_data(request_form_data_as_list), dict
        )
        document_mid: str = assert_cast(request_form_dict["document_mid"], str)
        context_document_mid: str = assert_cast(
            request_form_dict["context_document_mid"], str
        )

        # FIXME: Rewrite the legacy way of parsing by also use data from
        #        request_form_dict above.
        config_fields: Dict[str, List[str]] = defaultdict(list)
        for field_name, field_value in request_form_data.multi_items():
            result = re.search(r"^document\[(.*)]$", field_name)
            if result is not None:
                config_fields[result.group(1)].append(field_value)
        document_title: str = (
            config_fields["TITLE"][0] if "TITLE" in config_fields else ""
        )
        document_title = sanitize_html_form_field(
            document_title, multiline=False
        )
        document_title = document_title if document_title is not None else ""

        form_object = IncludedDocumentFormObject(
            document_mid=document_mid,
            context_document_mid=context_document_mid,
            document_title=document_title,
            jinja_environment=jinja_environment,
        )
        return form_object

    @staticmethod
    def create_from_document(
        *,
        document: SDocDocument,
        context_document_mid: str,
        jinja_environment: JinjaEnvironment,
    ) -> "IncludedDocumentFormObject":
        assert isinstance(document, SDocDocument)

        return IncludedDocumentFormObject(
            document_mid=document.reserved_mid,
            context_document_mid=context_document_mid,
            document_title=document.title,
            jinja_environment=jinja_environment,
        )

    def render_edit_form(self) -> Markup:
        rendered_template = self.jinja_environment.render_template_as_markup(
            "components/included_document_form/index.jinja", form_object=self
        )
        return render_turbo_stream(
            content=rendered_template,
            action="replace",
            target=f"article-{self.document_mid}",
        )



from dataclasses import dataclass
from typing import Iterator, Tuple, Union

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_from_file import DocumentFromFile
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentFromFileIF,
    SDocDocumentIF,
    SDocElementIF,
    SDocIteratedElementIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.node import (
    SDocNode,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.helpers.cast import assert_cast


@dataclass
class DocumentIterationContext:
    node: SDocElementIF
    level_stack: Tuple[int, ...]
    custom_level: bool = False

    def get_level(self) -> int:
        return len(self.level_stack)

    def get_level_string(self) -> str:
        if (
            isinstance(self.node, SDocNode)
            and self.node.node_type == "TEXT"
            and self.node.reserved_title is None
        ):
            return ""

        if self.node.ng_resolved_custom_level == "None":
            return ""

        if self.node.ng_resolved_custom_level is not None:
            return self.node.ng_resolved_custom_level

        if self.custom_level:
            return ""

        return ".".join(map(str, self.level_stack))


class DocumentCachingIterator:
    def __init__(self, document: SDocDocument) -> None:
        assert isinstance(document, SDocDocument), document

        self.document: SDocDocument = document

    def table_of_contents(
        self,
    ) -> Iterator[Tuple[SDocElementIF, DocumentIterationContext]]:
        for node_, context_ in self.all_content(
            print_fragments=True,
        ):
            if isinstance(node_, SDocNode):
                if node_.reserved_title is None:
                    continue
                if (
                    not self.document.config.is_requirement_in_toc()
                    and node_.node_type != "SECTION"
                ):
                    continue
            yield node_, context_

    def all_content(
        self,
        print_fragments: bool = False,
    ) -> Iterator[Tuple[SDocIteratedElementIF, DocumentIterationContext]]:
        root_node = self.document

        yield from self.all_node_content(
            root_node,
            print_fragments=print_fragments,
            level_stack=(),
            custom_level=not root_node.config.auto_levels,
        )

    def all_node_content(
        self,
        node: Union[SDocElementIF, DocumentFromFile],
        print_fragments: bool = False,
        level_stack: Tuple[int, ...] = (),
        custom_level: bool = False,
    ) -> Iterator[Tuple[SDocIteratedElementIF, DocumentIterationContext]]:
        if isinstance(node, SDocSection):
            # If node is not whitelisted, we ignore it. Also, note that due to
            # this early return, all child nodes of this node are ignored
            # as well because they are not added to the iteration queue.
            if not node.ng_whitelisted:
                return

            # FIXME: This will be changed to yield only context.
            context = DocumentIterationContext(
                node, level_stack, custom_level=custom_level
            )
            node.context.title_number_string = context.get_level_string()
            node.ng_level = context.get_level()

            yield node, context

            current_number = 0
            for subnode_ in node.section_contents:
                if subnode_.ng_resolved_custom_level is None and not (
                    isinstance(subnode_, SDocNode)
                    and subnode_.node_type == "TEXT"
                    and subnode_.reserved_title is None
                ):
                    current_number += 1

                yield from self.all_node_content(
                    assert_cast(
                        subnode_,
                        (
                            SDocNodeIF,
                            SDocSectionIF,
                            SDocDocumentIF,
                            SDocDocumentFromFileIF,
                        ),
                    ),
                    print_fragments=print_fragments,
                    level_stack=level_stack + (current_number,),
                    custom_level=custom_level
                    or subnode_.ng_resolved_custom_level is not None,
                )

        elif isinstance(node, SDocNode):
            # If node is not whitelisted, we ignore it. Also, note that due to
            # this early return, all child nodes of this node are ignored
            # as well because they are not added to the iteration queue.
            if not node.ng_whitelisted:
                return

            # FIXME: This will be changed to yield only context.
            context = DocumentIterationContext(
                node, level_stack, custom_level=custom_level
            )
            node.context.title_number_string = context.get_level_string()
            node.ng_level = context.get_level()

            yield node, context

            current_number = 0
            if node.section_contents is not None:
                for subnode_ in node.section_contents:
                    if subnode_.ng_resolved_custom_level is None and not (
                        isinstance(subnode_, SDocNode)
                        and subnode_.node_type == "TEXT"
                    ):
                        current_number += 1

                    yield from self.all_node_content(
                        subnode_,
                        print_fragments=print_fragments,
                        level_stack=level_stack + (current_number,),
                        custom_level=custom_level
                        or subnode_.ng_resolved_custom_level is not None,
                    )

        elif isinstance(node, SDocDocument):
            if (
                print_fragments
                and node.document_is_included()
                and self.document != node
            ):
                context = DocumentIterationContext(
                    node, level_stack, custom_level=custom_level
                )
                node.context.title_number_string = context.get_level_string()
                node.ng_level = context.get_level()

                yield node, context

            current_number = 0
            for subnode_ in node.section_contents:
                if subnode_.ng_resolved_custom_level is None and not (
                    isinstance(subnode_, SDocNode)
                    and subnode_.node_type == "TEXT"
                ):
                    current_number += 1
                yield from self.all_node_content(
                    assert_cast(
                        subnode_,
                        (
                            SDocNodeIF,
                            SDocSectionIF,
                            SDocDocumentIF,
                            SDocDocumentFromFileIF,
                        ),
                    ),
                    print_fragments=print_fragments,
                    level_stack=level_stack + (current_number,),
                    custom_level=custom_level
                    or subnode_.ng_resolved_custom_level is not None,
                )

        elif isinstance(node, DocumentFromFile):
            if not print_fragments:
                return

            assert node.resolved_document is not None

            yield from self.all_node_content(
                node.resolved_document,
                print_fragments=print_fragments,
                level_stack=level_stack,
            )

        else:
            raise NotImplementedError



import uuid


class MID(str):
    def __new__(cls, mid_value: str) -> "MID":
        assert isinstance(mid_value, str) and len(mid_value) > 0, mid_value
        return super().__new__(cls, mid_value)

    @staticmethod
    def create() -> "MID":
        return MID(uuid.uuid4().hex)

    def get_string_value(self) -> str:
        return str(self)
